# encoding: utf-8
#
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http:# mozilla.org/MPL/2.0/.
#
# Contact: Kyle Lahnakoski (kyle@lahnakoski.com)
#
from __future__ import absolute_import, division, unicode_literals

from jx_base.expressions import FromUnixOp as FromUnixOp_
from jx_bigquery.expressions._utils import check
from jx_bigquery.expressions.sql_script import SQLScript
from jx_bigquery.sql import sql_call
from mo_json import TIME


class FromUnixOp(FromUnixOp_):
    @check
    def to_bq(self, schema, not_null=False, boolean=False):
        v = self.value.to_bq(schema)[0].sql.n

        output = SQLScript(
            data_type=TIME,
            expr=sql_call("FROM_UNIXTIME", v),
            frum=self,
            miss=self.missing(),
            many=False,
            schema=schema,
        )
        return output



# encoding: utf-8
#
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http:# mozilla.org/MPL/2.0/.
#
# Contact: Kyle Lahnakoski (kyle@lahnakoski.com)
#
from __future__ import absolute_import, division, unicode_literals

from jx_base.expressions import NotRightOp as NotRightOp_
from jx_bigquery.expressions._utils import check
from mo_dots import wrap
from mo_sql import sql_iso


class NotRightOp(NotRightOp_):
    @check
    def to_bq(self, schema, not_null=False, boolean=False):
        v = self.value.to_bq(schema, not_null=True)[0].sql.s
        r = self.length.to_bq(schema, not_null=True)[0].sql.n
        l = "max(0, length" + sql_iso(v) + "-max(0, " + r + "))"
        expr = "SUBSTR" + sql_iso(v + ", 1, " + l)
        return wrap([{"name": ".", "sql": {"s": expr}}])



# encoding: utf-8
#
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http:# mozilla.org/MPL/2.0/.
#
# Contact: Kyle Lahnakoski (kyle@lahnakoski.com)
#
from __future__ import absolute_import, division, unicode_literals

from jx_base.expressions import MaxOp as MaxOp_
from jx_bigquery.expressions._utils import BQLang, check
from mo_dots import wrap
from mo_sql import sql_iso, sql_list


class MaxOp(MaxOp_):
    @check
    def to_bq(self, schema, not_null=False, boolean=False):
        terms = [BQLang[t].partial_eval().to_bq(schema)[0].sql.n for t in self.terms]
        return wrap([{"name": ".", "sql": {"n": "max" + sql_iso((sql_list(terms)))}}])



# encoding: utf-8
#
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http:# mozilla.org/MPL/2.0/.
#
# Contact: Kyle Lahnakoski (kyle@lahnakoski.com)
#
from __future__ import absolute_import, division, unicode_literals

from jx_base.expressions import ExistsOp as ExistsOp_
from jx_bigquery.expressions._utils import check
from mo_dots import wrap
from mo_sql import SQL_FALSE, SQL_IS_NOT_NULL, SQL_OR, sql_iso


class ExistsOp(ExistsOp_):
    @check
    def to_bq(self, schema, not_null=False, boolean=False):
        field = self.field.to_bq(schema)[0].sql
        acc = []
        for t, v in field.items():
            if t in "bns":
                acc.append(sql_iso(v + SQL_IS_NOT_NULL))

        if not acc:
            return wrap([{"name": ".", "sql": {"b": SQL_FALSE}}])
        else:
            return wrap([{"name": ".", "sql": {"b": SQL_OR.join(acc)}}])



# encoding: utf-8
#
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.
#

from __future__ import absolute_import, division, unicode_literals

import string
from collections import Mapping

from jx_python import jx
from mo_dots import listwrap, wrap
from mo_files.url import hex2chr
from mo_future import text, first, is_text
from mo_logs import Log
from mo_times import Date, Duration
from mo_sql import (
    SQL,
    SQL_FALSE,
    SQL_NULL,
    SQL_SELECT,
    SQL_TRUE,
    sql_iso,
    sql_list,
    SQL_AND,
    ConcatSQL,
    SQL_EQ,
    SQL_IS_NULL,
    SQL_COMMA,
    JoinSQL,
    SQL_FROM,
    SQL_WHERE,
    SQL_ORDERBY,
    SQL_STAR,
    SQL_LT,
    SQL_AS,
)
from mo_times.dates import parse

TIMESTAMP_FORMAT = "%Y-%m-%d %H:%M:%S.%f"
ALLOWED = string.ascii_letters + string.digits
GUID = "_id"  # user accessible, unique value across many machines
UID = "__id__"  # internal numeric id for single-database use


def quote_column(name):
    if not isinstance(name, ApiName):
        Log.error("expecting ApiName")
    esc_name = SQL(".".join("`" + n + "`" for n in name.values))
    return esc_name


class ApiName(object):
    """
    REPRESENT NAMES FROM/TO THE BIGQUERY API
    All names from the API should be wrapped with this
    All names being sent to API should be of this type (name == text(ApiName(name))
    """

    __slots__ = ["values"]

    def __init__(self, *values):
        """
        :param values:  DOt-delimited API names
        """
        if any(not is_text(n) for n in values):
            Log.error("expecting strings")
        self.values = values

    def __add__(self, other):
        if not isinstance(other, ApiName):
            Log.error("not alloweed")
        return ApiName(*(self.values + other.values))

    def __radd__(self, other):
        Log.error("disabled")

    def __iadd__(self, other):
        Log.error("disabled")

    def __eq__(self, other):
        if other == None:
            return False
        if not isinstance(other, ApiName):
            Log.error("not alloweed")
        return self.values == other.values

    def __str__(self):
        return ".".join(self.values)

    def __bool__(self):
        return bool(self.values)


def escape_name(name):
    if isinstance(name, ApiName):
        return name

    def quote(c):
        if c == "_":
            return "__"
        if c in ALLOWED:
            return c
        return "_" + hex(ord(c))[2:] + "_"

    esc_name = "".join(map(quote, name))
    return ApiName(esc_name)


def unescape_name(esc_name):
    if not isinstance(esc_name, ApiName):
        Log.error("expecting an api name")
    if len(esc_name.values) > 1:
        Log.error("do not knwo how to handle")
    try:
        parts = text(esc_name).split("_")
        result = parts[:1]
        for i, (p, q) in jx.chunk(parts[1:], 2):
            if len(p) == 0:
                result.append("_")
            else:
                result.append(hex2chr(p))
            result.append(q)
        name = "".join(result)
        return name
    except Exception:
        return esc_name


def sql_alias(value, alias):
    if not isinstance(value, SQL) or not isinstance(alias, ApiName):
        Log.error("Expecting (SQL, ApiName) parameters")
    return ConcatSQL(value, SQL_AS, quote_column(alias))


def sql_call(func_name, *parameters):
    return ConcatSQL(SQL(func_name), sql_iso(JoinSQL(SQL_COMMA, parameters)))


def quote_value(value):
    if isinstance(value, (Mapping, list)):
        return SQL(".")
    elif isinstance(value, Date):
        return quote_value(value.format(TIMESTAMP_FORMAT))
    elif isinstance(value, Duration):
        return SQL(text(value.seconds))
    elif is_text(value):
        return SQL("'" + value.replace("'", "''") + "'")
    elif value == None:
        return SQL_NULL
    elif value is True:
        return SQL_TRUE
    elif value is False:
        return SQL_FALSE
    else:
        return SQL(text(value))


def quote_list(values):
    return sql_iso(sql_list(map(quote_value, values)))


def sql_eq(**item):
    """
    RETURN SQL FOR COMPARING VARIABLES TO VALUES (AND'ED TOGETHER)

    :param item: keyword parameters representing variable and value
    :return: SQL
    """
    return SQL_AND.join(
        [
            ConcatSQL(quote_column(k), SQL_EQ, quote_value(v))
            if v != None
            else ConcatSQL(quote_column(k), SQL_IS_NULL)
            for k, v in item.items()
        ]
    )


def sql_lt(**item):
    """
    RETURN SQL FOR LESS-THAN (<) COMPARISION BETWEEN VARIABLES TO VALUES

    :param item: keyword parameters representing variable and value
    :return: SQL
    """
    k, v = first(item.items())
    return ConcatSQL(quote_column(k), SQL_LT, quote_value(v))


def sql_query(command):
    """
    VERY BASIC QUERY EXPRESSION TO SQL
    :param command: jx-expression
    :return: SQL
    """
    command = wrap(command)
    acc = [SQL_SELECT]
    if command.select:
        acc.append(JoinSQL(SQL_COMMA, map(quote_column, listwrap(command.select))))
    else:
        acc.append(SQL_STAR)

    acc.append(SQL_FROM)
    acc.append(quote_column(command["from"]))
    if command.where.eq:
        acc.append(SQL_WHERE)
        acc.append(sql_eq(**command.where.eq))
    if command.orderby:
        acc.append(SQL_ORDERBY)
        acc.append(JoinSQL(SQL_COMMA, map(quote_column, listwrap(command.orderby))))
    return ConcatSQL(*acc)



# encoding: utf-8
#
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http:# mozilla.org/MPL/2.0/.
#
# Contact: Kyle Lahnakoski (kyle@lahnakoski.com)
#
from __future__ import absolute_import, division, unicode_literals

from jx_base.expressions import BasicSubstringOp as BasicSubstringOp_
from jx_bigquery.expressions._utils import BQLang, check
from jx_bigquery.expressions.add_op import AddOp
from jx_bigquery.expressions.literal import Literal
from jx_bigquery.expressions.sub_op import SubOp
from mo_dots import wrap
from mo_sql import sql_iso


class BasicSubstringOp(BasicSubstringOp_):
    @check
    def to_bq(self, schema, not_null=False, boolean=False):
        value = BQLang[self.value].to_bq(schema, not_null=True)[0].sql.s
        start = (
            AddOp([self.start, Literal(1)])
            .partial_eval()
            .to_bq(schema, not_null=True)[0]
            .sql.n
        )
        length = (
            SubOp([self.end, self.start])
            .partial_eval()
            .to_bq(schema, not_null=True)[0]
            .sql.n
        )
        sql = "SUBSTR" + sql_iso(value + "," + start + ", " + length)
        return wrap([{"name": ".", "sql": {"s": sql}}])



# encoding: utf-8
#
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http:# mozilla.org/MPL/2.0/.
#
# Contact: Kyle Lahnakoski (kyle@lahnakoski.com)
#
from __future__ import absolute_import, division, unicode_literals

from jx_base.expressions import FindOp as FindOp_, ZERO, simplified
from jx_bigquery.expressions._utils import BQLang, check
from jx_bigquery.expressions.and_op import AndOp
from jx_bigquery.expressions.eq_op import EqOp
from jx_bigquery.expressions.not_left_op import NotLeftOp
from jx_bigquery.expressions.not_right_op import NotRightOp
from jx_bigquery.expressions.or_op import OrOp
from jx_bigquery.expressions.sql_instr_op import SqlInstrOp
from mo_dots import coalesce, wrap
from mo_sql import (
    SQL,
    SQL_CASE,
    SQL_ELSE,
    SQL_END,
    SQL_NULL,
    SQL_THEN,
    SQL_WHEN,
    sql_iso,
    sql_list,
    SQL_ZERO,
)


class FindOp(FindOp_):
    @simplified
    def partial_eval(self):
        return FindOp(
            [BQLang[self.value].partial_eval(), BQLang[self.find].partial_eval()],
            **{
                "start": BQLang[self.start].partial_eval(),
                "default": BQLang[self.default].partial_eval(),
            }
        )

    @check
    def to_bq(self, schema, not_null=False, boolean=False):
        value = BQLang[self.value].partial_eval().to_bq(schema)[0].sql.s
        find = BQLang[self.find].partial_eval().to_bq(schema)[0].sql.s
        start = BQLang[self.start].partial_eval().to_bq(schema)[0].sql.n
        default = coalesce(
            BQLang[self.default].partial_eval().to_bq(schema)[0].sql.n, SQL_NULL
        )

        if start.sql != SQL_ZERO.sql.strip():
            value = NotRightOp([self.value, self.start]).to_bq(schema)[0].sql.s

        index = "INSTR" + sql_iso(sql_list([value, find]))

        sql = (
            SQL_CASE
            + SQL_WHEN
            + index
            + SQL_THEN
            + index
            + SQL("-1+")
            + start
            + SQL_ELSE
            + default
            + SQL_END
        )

        return wrap([{"name": ".", "sql": {"n": sql}}])

    def exists(self):
        output = OrOp(
            [
                self.default.exists(),
                AndOp(
                    [
                        self.value.exists(),
                        self.find.exists(),
                        EqOp(
                            [
                                SqlInstrOp(
                                    [NotLeftOp([self.value, self.start]), self.find]
                                ),
                                ZERO,
                            ]
                        ),
                    ]
                ),
            ]
        )
        return output



# encoding: utf-8
#
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http:# mozilla.org/MPL/2.0/.
#
# Contact: Kyle Lahnakoski (kyle@lahnakoski.com)
#
from __future__ import absolute_import, division, unicode_literals

from jx_base.expressions import (
    EqOp as EqOp_,
    FALSE,
    TRUE,
    ZERO,
    simplified,
)
from jx_base.expressions._utils import builtin_ops
from jx_bigquery.expressions._utils import BQLang, check
from jx_bigquery.expressions.case_op import CaseOp
from jx_bigquery.expressions.literal import Literal
from jx_bigquery.expressions.sql_eq_op import SqlEqOp
from jx_bigquery.expressions.bql_script import BQLScript
from jx_bigquery.expressions.when_op import WhenOp
from mo_json import BOOLEAN
from mo_logs import Log
from mo_sql import SQL_FALSE, SQL_IS_NULL, SQL_OR, sql_iso


class EqOp(EqOp_):
    @check
    def to_bq(self, schema, not_null=False, boolean=False):
        lhs = BQLang[self.lhs].to_bq(schema)
        rhs = BQLang[self.rhs].to_bq(schema)
        acc = []
        if len(lhs) != len(rhs):
            Log.error("lhs and rhs have different dimensionality!?")

        for l, r in zip(lhs, rhs):
            for t in "bsnj":
                if l.sql[t] == None:
                    if r.sql[t] == None:
                        pass
                    else:
                        acc.append(sql_iso(r.sql[t]) + SQL_IS_NULL)
                elif l.sql[t] is ZERO:
                    if r.sql[t] == None:
                        acc.append(SQL_FALSE)
                    elif r.sql[t] is ZERO:
                        Log.error(
                            "Expecting expression to have been simplified already"
                        )
                    else:
                        acc.append(r.sql[t])
                else:
                    if r.sql[t] == None:
                        acc.append(sql_iso(l.sql[t]) + SQL_IS_NULL)
                    elif r.sql[t] is ZERO:
                        acc.append(l.sql[t])
                    else:
                        acc.append(sql_iso(l.sql[t]) + " = " + sql_iso(r.sql[t]))
        if not acc:
            return FALSE.to_bq(schema)
        else:
            return BQLScript(
                expr=SQL_OR.join(acc),
                frum=self,
                data_type=BOOLEAN,
                miss=FALSE,
                schema=schema,
            )

    @simplified
    def partial_eval(self):
        lhs = self.lhs.partial_eval()
        rhs = self.rhs.partial_eval()

        if isinstance(lhs, Literal) and isinstance(rhs, Literal):
            return TRUE if builtin_ops["eq"](lhs.value, rhs.value) else FALSE
        else:
            rhs_missing = rhs.missing().partial_eval()
            output = CaseOp(
                [
                    WhenOp(lhs.missing(), **{"then": rhs_missing}),
                    WhenOp(rhs_missing, **{"then": FALSE}),
                    SqlEqOp([lhs, rhs]),
                ]
            ).partial_eval()
            return output



# encoding: utf-8
#
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http:# mozilla.org/MPL/2.0/.
#
# Contact: Kyle Lahnakoski (kyle@lahnakoski.com)
#
from __future__ import absolute_import, division, unicode_literals

from jx_base.expressions import LeavesOp as LeavesOp_
from jx_base.language import is_op
from jx_bigquery.expressions._utils import check
from jx_bigquery.expressions.variable import Variable
from mo_dots import join_field, split_field, startswith_field, wrap
from mo_json import EXISTS, NESTED, OBJECT
from mo_logs import Log


class LeavesOp(LeavesOp_):
    @check
    def to_bq(self, schema, not_null=False, boolean=False):
        if not is_op(self.term, Variable):
            Log.error("Can only handle Variable")
        term = self.term.var
        prefix_length = len(split_field(term))
        output = wrap(
            [
                {
                    "name": join_field(
                        split_field(schema.get_column_name(c))[prefix_length:]
                    ),
                    "sql": Variable(schema.get_column_name(c)).to_bq(schema)[0].sql,
                }
                for c in schema.columns
                if startswith_field(c.name, term)
                and (
                    (
                        c.jx_type not in (EXISTS, OBJECT, NESTED)
                        and startswith_field(schema.nested_path[0], c.nested_path[0])
                    )
                    or (
                        c.jx_type not in (EXISTS, OBJECT)
                        and schema.nested_path[0] == c.nested_path[0]
                    )
                )
            ]
        )
        return output



# encoding: utf-8
#
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http:# mozilla.org/MPL/2.0/.
#
# Contact: Kyle Lahnakoski (kyle@lahnakoski.com)
#
from __future__ import absolute_import, division, unicode_literals

from google.cloud import bigquery
from google.cloud.bigquery import TimePartitioning
from mo_dots import Null
from mo_future import first
from mo_kwargs import override
from mo_logs import Log
from mo_times import Duration, DAY, YEAR

NEVER = 10 * YEAR


class Partition(object):
    """
    DESCRIBE HOW TO PARTITION TABLE
    """

    __slots__ = ["field", "interval", "expire"]

    @override
    def __new__(cls, field=None, interval=DAY, expire=NEVER, flake=Null, kwargs=None):
        if field == None:
            return Null
        return object.__new__(cls)

    @override
    def __init__(self, field, interval=DAY, expire=NEVER, flake=Null, kwargs=None):
        column = first(flake.leaves(field))
        if not column:
            Log.error("expecting {{field}} in snowflake for partitioning", field=field)

        self.field = column.es_column
        self.interval = Duration(interval)
        self.expire = Duration(expire)
        if not isinstance(self.interval, Duration) or not isinstance(
            self.expire, Duration
        ):
            Log.error("expecting durations")

    @property
    def bq_time_partitioning(self):
        return TimePartitioning(
            type_=bigquery.TimePartitioningType.DAY,
            field=self.field,
            expiration_ms=int(self.expire.milli),
        )



# encoding: utf-8
#
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http:# mozilla.org/MPL/2.0/.
#
# Contact: Kyle Lahnakoski (kyle@lahnakoski.com)
#
from __future__ import absolute_import, division, unicode_literals

from jx_base.expressions import ConcatOp as ConcatOp_, TrueOp
from jx_bigquery.expressions._utils import BQLang, check
from jx_bigquery.expressions.length_op import LengthOp
from jx_bigquery.expressions.bql_script import BQLScript
from mo_dots import coalesce
from mo_json import STRING
from mo_sql import (
    SQL,
    SQL_CASE,
    SQL_ELSE,
    SQL_EMPTY_STRING,
    SQL_END,
    SQL_NULL,
    SQL_THEN,
    SQL_WHEN,
    sql_iso,
    sql_list,
    sql_concat_text,
)
from jx_mysql.mysql import quote_value


class ConcatOp(ConcatOp_):
    @check
    def to_bq(self, schema, not_null=False, boolean=False):
        default = self.default.to_bq(schema)
        if len(self.terms) == 0:
            return default
        default = coalesce(default[0].sql.s, SQL_NULL)
        sep = BQLang[self.separator].to_bq(schema)[0].sql.s

        acc = []
        for t in self.terms:
            t = BQLang[t]
            missing = t.missing().partial_eval()

            term = t.to_bq(schema, not_null=True)[0].sql
            if term.s:
                term_sql = term.s
            elif term.n:
                term_sql = "cast(" + term.n + " as text)"
            else:
                term_sql = (
                    SQL_CASE
                    + SQL_WHEN
                    + term.b
                    + SQL_THEN
                    + quote_value("true")
                    + SQL_ELSE
                    + quote_value("false")
                    + SQL_END
                )

            if isinstance(missing, TrueOp):
                acc.append(SQL_EMPTY_STRING)
            elif missing:
                acc.append(
                    SQL_CASE
                    + SQL_WHEN
                    + sql_iso(missing.to_bq(schema, boolean=True)[0].sql.b)
                    + SQL_THEN
                    + SQL_EMPTY_STRING
                    + SQL_ELSE
                    + sql_iso(sql_concat_text([sep, term_sql]))
                    + SQL_END
                )
            else:
                acc.append(sql_concat_text([sep, term_sql]))

        expr_ = "SUBSTR" + sql_iso(
            sql_list(
                [
                    sql_concat_text(acc),
                    LengthOp(self.separator).to_bq(schema)[0].sql.n + SQL("+1"),
                ]
            )
        )

        return BQLScript(
            expr=expr_,
            data_type=STRING,
            frum=self,
            miss=self.missing(),
            many=False,
            schema=schema,
        )



# encoding: utf-8
#
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http:# mozilla.org/MPL/2.0/.
#
# Contact: Kyle Lahnakoski (kyle@lahnakoski.com)
#
from __future__ import absolute_import, division, unicode_literals

from jx_base.expressions import BooleanOp as BooleanOp_, FALSE, TRUE, is_literal
from jx_bigquery.expressions._utils import BQLang, check


class BooleanOp(BooleanOp_):
    @check
    def to_bq(self, schema, not_null=False, boolean=False):
        term = BQLang[self.term].partial_eval()
        if term.type == "boolean":
            sql = term.to_bq(schema)
            return sql
        elif is_literal(term) and term.value in ("T", "F"):
            if term.value == "T":
                return TRUE.to_bq(schema)
            else:
                return FALSE.to_bq(schema)
        else:
            sql = term.exists().partial_eval().to_bq(schema)
            return sql



# encoding: utf-8
#
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http:# mozilla.org/MPL/2.0/.
#
# Contact: Kyle Lahnakoski (kyle@lahnakoski.com)
#
from __future__ import absolute_import, division, unicode_literals

from jx_bigquery.sql import quote_value

from jx_base.expressions import Literal as Literal_
from jx_bigquery.expressions._utils import check
from mo_dots import wrap
from mo_future import text
from mo_math import is_number


class Literal(Literal_):
    @check
    def to_bq(self, schema, not_null=False, boolean=False):
        value = self.value
        if value == None:
            return wrap([{"name": "."}])
        elif isinstance(value, text):
            return wrap([{"name": ".", "sql": {"s": quote_value(value)}}])
        elif is_number(value):
            return wrap([{"name": ".", "sql": {"n": quote_value(value)}}])
        elif value in [True, False]:
            return wrap([{"name": ".", "sql": {"b": quote_value(value)}}])
        else:
            return wrap([{"name": ".", "sql": {"j": quote_value(self.json)}}])



# encoding: utf-8
#
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http:# mozilla.org/MPL/2.0/.
#
# Contact: Kyle Lahnakoski (kyle@lahnakoski.com)
#
from __future__ import absolute_import, division, unicode_literals

from jx_base.expressions import IntegerOp as IntegerOp_
from jx_bigquery.expressions._utils import check
from mo_dots import wrap
from mo_sql import sql_coalesce


class IntegerOp(IntegerOp_):
    @check
    def to_bq(self, schema, not_null=False, boolean=False):
        value = self.term.to_bq(schema, not_null=True)
        acc = []
        for c in value:
            for t, v in c.sql.items():
                if t == "s":
                    acc.append("CAST(" + v + " as INTEGER)")
                else:
                    acc.append(v)

        if not acc:
            return wrap([])
        elif len(acc) == 1:
            return wrap([{"name": ".", "sql": {"n": acc[0]}}])
        else:
            return wrap([{"name": ".", "sql": {"n": sql_coalesce(acc)}}])



# encoding: utf-8
#
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http:# mozilla.org/MPL/2.0/.
#
# Contact: Kyle Lahnakoski (kyle@lahnakoski.com)
#
from __future__ import absolute_import, division, unicode_literals

import re
from copy import copy

from google.cloud import bigquery
from google.oauth2 import service_account
from jx_base import Container, Facts
from jx_python import jx
from mo_dots import listwrap, unwrap, join_field, Null, is_data, Data, wrap, set_default
from mo_future import is_text, text, first
from mo_json import INTEGER
from mo_kwargs import override
from mo_logs import Log, Except
from mo_math.randoms import Random
from mo_sql import (
    ConcatSQL,
    SQL,
    SQL_SELECT,
    JoinSQL,
    SQL_NULL,
    SQL_FROM,
    SQL_COMMA,
    SQL_AS,
    SQL_ORDERBY,
    SQL_CR,
    SQL_SELECT_AS_STRUCT,
    SQL_INSERT,
    SQL_DESC,
    SQL_UNION_ALL,
)
from mo_threads import Till
from mo_times import MINUTE, Timer
from mo_times.dates import Date

from jx_bigquery import snowflakes
from jx_bigquery.snowflakes import Snowflake
from jx_bigquery.sql import (
    quote_column,
    ALLOWED,
    sql_call,
    sql_alias,
    escape_name,
    ApiName,
    sql_query,
)
from jx_bigquery.typed_encoder import (
    NESTED_TYPE,
    typed_encode,
    REPEATED,
    json_type_to_bq_type,
    INTEGER_TYPE,
    untyped,
)

DEBUG = False
EXTEND_LIMIT = 2 * MINUTE  # EMIT ERROR IF ADDING RECORDS TO TABLE TOO OFTEN
MAX_MERGE = 10  # MAXIMUM NUMBER OF TABLES TO MERGE AT ONCE
SUFFIX_PATTERN = re.compile(r"__\w{20}")


def connect(account_info):
    creds = service_account.Credentials.from_service_account_info(info=account_info)
    client = bigquery.Client(project=account_info.project_id, credentials=creds)
    return client


def create_dataset(project_id, dataset, client):
    full_name = ApiName(project_id) + escape_name(dataset)

    _dataset = bigquery.Dataset(text(full_name))
    _dataset.location = "US"
    return client.create_dataset(_dataset)


def find_dataset(dataset, client):
    esc_name = escape_name(dataset)

    datasets = list(client.list_datasets())
    for _dataset in datasets:
        if ApiName(_dataset.dataset_id) == esc_name:
            return _dataset.reference


class Dataset(Container):
    """
    REPRESENT A BIGQUERY DATASET; aka A CONTAINER FOR TABLES; aka A DATABASE
    """

    @override
    def __init__(self, dataset, account_info, kwargs):
        self.client = connect(account_info)
        self.short_name = dataset
        esc_name = escape_name(dataset)
        self.full_name = ApiName(account_info.project_id) + esc_name

        self.dataset = find_dataset(dataset, self.client)
        if not self.dataset:
            self.dataset = create_dataset(account_info.project_id, dataset, self.client)

    @override
    def get_or_create_table(
        self,
        table,
        schema=None,
        typed=True,
        read_only=False,
        sharded=False,
        partition=None,
        cluster=None,  # TUPLE OF FIELDS TO SORT DATA
        id=None,
        kwargs=None,
    ):
        if kwargs.lookup != None or kwargs.flake != None:
            Log.error("expecting schema, not lookup")
        try:
            return Table(kwargs=kwargs, container=self)
        except Exception as e:
            e = Except.wrap(e)
            if "Not found: Table" in e:
                return self.create_table(kwargs)
            Log.error("could not get table {{table}}", table=table, cause=e)

    @override
    def create_or_replace_table(
        self,
        table,
        schema=None,
        typed=True,
        read_only=False,
        partition=None,
        cluster=None,  # TUPLE OF FIELDS TO SORT DATA
        sharded=False,
        kwargs=None,
    ):
        if kwargs.lookup != None or kwargs.flake != None:
            Log.error("expecting schema, not lookup")

        try:
            self.delete_table(table)
        except Exception as e:
            e = Except.wrap(e)
            if "Not found: Table" not in e and "Unable to get TableReference" not in e:
                Log.error("could not get table {{table}}", table=table, cause=e)
        return self.create_table(kwargs=kwargs)

    def delete_table(self, name):
        api_name = escape_name(name)

        tables = list(self.client.list_tables(self.dataset))
        for table_item in tables:
            table = table_item.reference
            table_api_name = ApiName(table.table_id)
            if text(table_api_name).startswith(text(api_name)):
                if table_api_name == api_name:
                    if table_item.table_type != "VIEW":
                        Log.error("expecting {{table}} to be a view", table=api_name)
                    self.client.delete_table(table)
                elif SUFFIX_PATTERN.match(text(table_api_name)[len(text(api_name)) :]):
                    try:
                        self.client.delete_table(table)
                    except Exception as e:
                        Log.warning(
                            "could not delete table {{table}}", table=table, cause=e
                        )

    @override
    def create_table(
        self,
        table,
        schema=None,
        typed=True,
        read_only=True,  # TO PREVENT ACCIDENTAL WRITING
        sharded=False,
        partition=Null,  # PARTITION RULES
        cluster=None,  # TUPLE OF FIELDS TO SORT DATA
        top_level_fields=Null,
        kwargs=None,
    ):
        if kwargs.lookup != None or kwargs.flake != None:
            Log.error("expecting schema, not lookup")
        full_name = self.full_name + escape_name(table)
        if not schema:
            # WE MUST HAVE SOMETHING
            if typed:
                schema = copy(DEFAULT_TYPED_SCHEMA)
            else:
                schema = copy(DEFAULT_SCHEMA)

        flake = Snowflake(text(full_name), top_level_fields, partition, schema=schema)

        if read_only:
            Log.error("Can not create a table for read-only use")

        if sharded:
            shard_name = escape_name(table + "_" + "".join(Random.sample(ALLOWED, 20)))
            shard_api_name = self.full_name + shard_name
            _shard = bigquery.Table(text(shard_api_name), schema=flake.to_bq_schema())
            _shard.time_partitioning = unwrap(flake._partition.bq_time_partitioning)
            _shard.clustering_fields = [
                c.es_column
                for f in listwrap(cluster)
                for c in [first(flake.leaves(f))]
                if c
            ] or None
            self.shard = self.client.create_table(_shard)
            self.create_view(full_name, shard_api_name)
        else:
            _table = bigquery.Table(text(full_name), schema=flake.to_bq_schema())
            _table.time_partitioning = unwrap(flake._partition.bq_time_partitioning)
            _table.clustering_fields = [
                l.es_column for f in listwrap(cluster) for l in flake.leaves(f)
            ] or None
            self.client.create_table(_table)
            Log.note("created table {{table}}", table=_table.table_id)

        return Table(
            table=table,
            typed=typed,
            read_only=read_only,
            sharded=sharded,
            partition=partition,
            top_level_fields=top_level_fields,
            kwargs=kwargs,
            container=self,
        )

    def create_view(self, view_api_name, shard_api_name):
        sql = ConcatSQL(
            SQL("CREATE VIEW\n"),
            quote_column(view_api_name),
            SQL_AS,
            sql_query({"from": shard_api_name}),
        )
        job = self.query_and_wait(sql)
        if job.errors:
            Log.error(
                "Can not create view\n{{sql}}\n{{errors|json|indent}}",
                sql=sql,
                errors=job.errors,
            )
        pass

    def query_and_wait(self, sql):
        job = self.client.query(text(sql))
        while job.state == "RUNNING":
            DEBUG and Log.note(
                "job {{id}} state = {{state}}", id=job.job_id, state=job.state
            )
            Till(seconds=1).wait()
            job = self.client.get_job(job.job_id)
        DEBUG and Log.note(
            "job {{id}} state = {{state}}", id=job.job_id, state=job.state
        )
        return job


class Table(Facts):
    @override
    def __init__(
        self,
        table,
        typed,
        read_only,
        sharded,
        container,
        id=Null,
        partition=Null,
        cluster=Null,
        top_level_fields=Null,
        kwargs=None,
    ):
        self.short_name = table
        self.typed = typed
        self.read_only = read_only
        self.cluster = cluster
        self.id = id
        self.top_level_fields = top_level_fields
        self.config = Data(  # USED TO REPLICATE THIS
            typed=typed,
            read_only=read_only,
            sharded=sharded,
            id=id,
            partition=partition,
            cluster=cluster,
            top_level_fields=top_level_fields,
        )

        esc_name = escape_name(table)
        self.full_name = container.full_name + esc_name
        self.alias_view = alias_view = container.client.get_table(text(self.full_name))
        self.partition = partition
        self.container = container

        if not sharded:
            if not read_only and alias_view.table_type == "VIEW":
                Log.error("Expecting a table, not a view")
            self.shard = alias_view
            self._flake = Snowflake.parse(
                alias_view.schema,
                text(self.full_name),
                self.top_level_fields,
                partition,
            )
        else:
            if alias_view.table_type != "VIEW":
                Log.error("Sharded tables require a view")
            current_view = container.client.get_table(text(self.full_name))
            view_sql = current_view.view_query
            shard_name = _extract_primary_shard_name(view_sql)
            try:
                self.shard = container.client.get_table(
                    text(container.full_name + shard_name)
                )
                self._flake = Snowflake.parse(
                    alias_view.schema,
                    text(self.full_name),
                    self.top_level_fields,
                    partition,
                )
            except Exception as e:
                Log.warning("view {{name}} is invalid", name=shard_name, cause=e)
                self._flake = Snowflake.parse(
                    alias_view.schema,
                    text(self.full_name),
                    self.top_level_fields,
                    partition,
                )
                # REMOVE STALE VIEW
                container.client.delete_table(current_view)

                # MAKE NEW VIEW POINTING TO NEW SHARD
                self._create_new_shard()
                container.create_view(
                    self.full_name,
                    self.container.full_name + ApiName(self.shard.table_id),
                )

        self.last_extend = Date.now() - EXTEND_LIMIT

    def all_records(self):
        """
        MOSTLY FOR TESTING, RETURN ALL RECORDS IN TABLE
        :return:
        """
        sql = sql_query({"from": self.full_name})
        query_job = self.container.client.query(text(sql))

        # WE WILL REACH INTO THE _flake, SINCE THIS IS THE FIRST PLACE WE ARE ACTUALLY PULLING RECORDS OUT
        # TODO: WITH MORE CODE THIS LOGIC GOES ELSEWHERE
        _ = self._flake.columns  # ENSURE schema HAS BEEN PROCESSED
        if not self._flake._top_level_fields.keys():
            for row in query_job:
                yield untyped(dict(row))
        else:
            top2deep = {
                name: path for path, name in self._flake._top_level_fields.items()
            }
            for row in query_job:
                output = {}
                doc = dict(row)
                # COPY ALL BUT TOP LEVEL FIELDS
                for k, v in doc.items():
                    deep = top2deep.get(k)
                    if deep is None:
                        output[k] = v
                # INSERT TOP LEVEL FIELDS
                reach = wrap(output)
                for k, p in top2deep.items():
                    reach[p] = doc[k]
                yield untyped(output)

    @property
    def flake(self):
        return self._flake

    def _create_new_shard(self):
        primary_shard = self.container.create_table(
            table=self.short_name + "_" + "".join(Random.sample(ALLOWED, 20)),
            sharded=False,
            schema=self._flake.schema,
            kwargs=self.config,
        )
        self.shard = primary_shard.shard

    def extend(self, rows):
        if self.read_only:
            Log.error("not for writing")
        if len(rows) == 0:
            return

        try:
            update = {}
            with Timer("encoding"):
                while True:
                    output = []
                    for rownum, row in enumerate(rows):
                        typed, more, add_nested = typed_encode(row, self.flake)
                        set_default(update, more)
                        if add_nested:
                            # row HAS NEW NESTED COLUMN!
                            # GO OVER THE rows AGAIN SO "RECORD" GET MAPPED TO "REPEATED"
                            DEBUG and Log.note("New nested documnet found, retrying")
                            break
                        output.append(typed)
                    else:
                        break

            if update or not self.shard:
                # BATCH HAS ADDITIONAL COLUMNS!!
                # WE CAN NOT USE THE EXISTING SHARD, MAKE A NEW ONE:
                self._create_new_shard()
                Log.note(
                    "added new shard with name: {{shard}}", shard=self.shard.table_id
                )
            with Timer("insert {{num}} rows to bq", param={"num": len(rows)}):
                failures = self.container.client.insert_rows_json(
                    self.shard,
                    json_rows=output,
                    row_ids=[None] * len(output),
                    skip_invalid_rows=False,
                    ignore_unknown_values=False,
                )
            if failures:
                if all(r == "stopped" for r in wrap(failures).errors.reason):
                    self._create_new_shard()
                    Log.note(
                        "STOPPED encountered: Added new shard with name: {{shard}}",
                        shard=self.shard.table_id,
                    )
                Log.error(
                    "Got {{num}} failures:\n{{failures|json}}",
                    num=len(failures),
                    failures=failures[:5],
                )
            else:
                self.last_extend = Date.now()
                Log.note("{{num}} rows added", num=len(output))
        except Exception as e:
            e = Except.wrap(e)
            if "Your client has issued a malformed or illegal request." in e:
                Log.error("big query complains about {{data|json}}", data=output)
            if len(rows) > 1 and "Request payload size exceeds the limit" in e:
                # TRY A SMALLER BATCH
                cut = len(rows) // 2
                self.extend(rows[:cut])
                self.extend(rows[cut:])
                return
            Log.error("Do not know how to handle", cause=e)

    def add(self, row):
        self.extend([row])

    def merge_shards(self):
        shards = []
        tables = list(self.container.client.list_tables(self.container.dataset))
        current_view = Null  # VIEW THAT POINTS TO PRIMARY SHARD
        primary_shard_name = None  # PRIMARY SHARD
        api_name = escape_name(self.short_name)

        for table_item in tables:
            table = table_item.reference
            table_api_name = ApiName(table.table_id)
            if text(table_api_name).startswith(text(api_name)):
                if table_api_name == api_name:
                    if table_item.table_type != "VIEW":
                        Log.error("expecting {{table}} to be a view", table=api_name)
                    current_view = self.container.client.get_table(table)
                    view_sql = current_view.view_query
                    primary_shard_name = _extract_primary_shard_name(view_sql)
                elif SUFFIX_PATTERN.match(text(table_api_name)[len(text(api_name)) :]):
                    try:
                        known_table = self.container.client.get_table(table)
                        shards.append(known_table)
                    except Exception as e:
                        Log.warning(
                            "could not merge table {{table}}", table=table, cause=e
                        )

        if not current_view:
            Log.error(
                "expecting {{table}} to be a view pointing to a table", table=api_name
            )

        shard_flakes = [
            Snowflake.parse(
                big_query_schema=shard.schema,
                es_index=text(self.container.full_name + ApiName(shard.table_id)),
                top_level_fields=self.top_level_fields,
                partition=self.partition,
            )
            for shard in shards
        ]
        total_flake = snowflakes.merge(
            shard_flakes,
            es_index=text(self.full_name),
            top_level_fields=self.top_level_fields,
            partition=self.partition,
        )

        for i, s in enumerate(shards):
            if ApiName(s.table_id) == primary_shard_name:
                if total_flake == shard_flakes[i]:
                    # USE THE CURRENT PRIMARY SHARD AS A DESTINATION
                    del shards[i]
                    del shard_flakes[i]
                    break
        else:
            name = self.short_name + "_" + "".join(Random.sample(ALLOWED, 20))
            primary_shard_name = escape_name(name)
            self.container.create_table(
                table=name,
                schema=total_flake.schema,
                sharded=False,
                read_only=False,
                kwargs=self.config,
            )

        primary_full_name = self.container.full_name + primary_shard_name

        selects = []
        for flake, table in zip(shard_flakes, shards):
            q = ConcatSQL(
                SQL_SELECT,
                JoinSQL(ConcatSQL(SQL_COMMA, SQL_CR), gen_select(total_flake, flake)),
                SQL_FROM,
                quote_column(ApiName(table.dataset_id, table.table_id)),
            )
            selects.append(q)

        Log.note("inserting into table {{table}}", table=text(primary_shard_name))
        matched = []
        unmatched = []
        for sel, shard, flake in zip(selects, shards, shard_flakes):
            if flake == total_flake:
                matched.append((sel, shard, flake))
            else:
                unmatched.append((sel, shard, flake))

        # EVERYTHING THAT IS IDENTICAL TO PRIMARY CAN BE MERGED WITH SIMPLE UNION ALL
        if matched:
            for g, merge_chunk in jx.chunk(matched, MAX_MERGE):
                command = ConcatSQL(
                    SQL_INSERT,
                    quote_column(primary_full_name),
                    JoinSQL(
                        SQL_UNION_ALL,
                        (
                            sql_query(
                                {
                                    "from": self.container.full_name
                                    + ApiName(shard.table_id)
                                }
                            )
                            for _, shard, _ in merge_chunk
                        ),
                    ),
                )
                DEBUG and Log.note("{{sql}}", sql=text(command))
                job = self.container.query_and_wait(command)
                Log.note("job {{id}} state = {{state}}", id=job.job_id, state=job.state)

                if job.errors:
                    Log.error(
                        "\n{{sql}}\nDid not fill table:\n{{reason|json|indent}}",
                        sql=command.sql,
                        reason=job.errors,
                    )
                for _, shard, _ in merge_chunk:
                    self.container.client.delete_table(shard)

        # ALL OTHER SCHEMAS MISMATCH
        for s, shard, _ in unmatched:
            try:
                command = ConcatSQL(SQL_INSERT, quote_column(primary_full_name), s)
                DEBUG and Log.note("{{sql}}", sql=text(command))
                job = self.container.query_and_wait(command)
                Log.note(
                    "from {{shard}}, job {{id}}, state {{state}}",
                    id=job.job_id,
                    shard=shard.table_id,
                    state=job.state,
                )

                if job.errors:
                    if all(
                        " does not have a schema." in m
                        for m in wrap(job.errors).message
                    ):
                        pass  # NOTHING TO DO
                    else:
                        Log.error(
                            "\n{{sql}}\nDid not fill table:\n{{reason|json|indent}}",
                            sql=command.sql,
                            reason=job.errors,
                        )

                self.container.client.delete_table(shard)
            except Exception as e:
                Log.warning("failure to merge {{shard}}", shard=shard, cause=e)

        # REMOVE OLD VIEW
        view_full_name = self.container.full_name + api_name
        if current_view:
            self.container.client.delete_table(current_view)

        # CREATE NEW VIEW
        self.container.create_view(view_full_name, primary_full_name)

    def condense(self):
        """
        :return:
        """
        # MAKE NEW SHARD
        partition = JoinSQL(
            SQL_COMMA,
            [
                quote_column(c.es_field)
                for f in listwrap(self.id.field)
                for c in self.flake.leaves(f)
            ],
        )
        order_by = JoinSQL(
            SQL_COMMA,
            [
                ConcatSQL(quote_column(c.es_field), SQL_DESC)
                for f in listwrap(self.id.version)
                for c in self.flake.leaves(f)
            ],
        )
        # WRAP WITH etl.timestamp BEST SELECTION

        self.container.query_and_wait(
            ConcatSQL(
                SQL(  # SOME KEYWORDS: ROWNUM RANK
                    "SELECT * EXCEPT (_rank) FROM (SELECT *, ROW_NUMBER() OVER (PARTITION BY "
                ),
                partition,
                SQL_ORDERBY,
                order_by,
                SQL(") AS _rank FROM "),
                quote_column(self.full_name),
                SQL(") a WHERE _rank=1"),
            )
        )


def _extract_primary_shard_name(view_sql):
    # TODO: USE REAL PARSER
    e = view_sql.lower().find("from ")
    return ApiName(view_sql[e + 5 :].strip().split(".")[-1].strip("`"))


def gen_select(total_flake, flake):
    """
    GENERATE SELECT CLAUSE
    :param total_flake:
    :param flake:
    :return:
    """

    def _gen_select(
        source_path, source_tops, source_flake, total_path, total_tops, total_flake
    ):
        if total_flake == source_flake and not total_tops:
            return [
                quote_column(source_path + escape_name(k))
                for k in jx.sort(total_flake.keys())
            ]

        if NESTED_TYPE in total_flake:
            # PROMOTE EVERYTHING TO REPEATED
            v = source_flake.get(NESTED_TYPE)
            t = total_flake.get(NESTED_TYPE)

            if not v:
                # CONVERT INNER OBJECT TO ARRAY OF ONE STRUCT
                inner = [
                    ConcatSQL(
                        SQL_SELECT_AS_STRUCT,
                        JoinSQL(
                            ConcatSQL(SQL_COMMA, SQL_CR),
                            _gen_select(
                                source_path,
                                Null,
                                source_flake,
                                total_path + REPEATED,
                                Null,
                                t,
                            ),
                        ),
                    )
                ]
            else:
                row_name = "row" + text(len(source_path.values))
                ord_name = "ordering" + text(len(source_path.values))
                inner = [
                    ConcatSQL(
                        SQL_SELECT_AS_STRUCT,
                        JoinSQL(
                            ConcatSQL(SQL_COMMA, SQL_CR),
                            _gen_select(
                                ApiName(row_name), Null, v, ApiName(row_name), Null, t
                            ),
                        ),
                        SQL_FROM,
                        sql_call("UNNEST", quote_column(source_path + REPEATED)),
                        SQL_AS,
                        SQL(row_name),
                        SQL(" WITH OFFSET AS "),
                        SQL(ord_name),
                        SQL_ORDERBY,
                        SQL(ord_name),
                    )
                ]

            return [sql_alias(sql_call("ARRAY", *inner), REPEATED)]

        selection = []
        for k, t in jx.sort(total_flake.items(), 0):
            k_total_tops = total_tops if is_text(total_tops) else total_tops[k]
            k_tops = source_tops if is_text(source_tops) else source_tops[k]
            v = source_flake.get(k)
            if is_text(k_total_tops):
                # DO NOT INCLUDE TOP_LEVEL_FIELDS
                pass
            elif t == v and not k_total_tops and not k_tops:
                selection.append(
                    ConcatSQL(
                        quote_column(source_path + escape_name(k)),
                        SQL_AS,
                        quote_column(escape_name(k)),
                    )
                )
            elif is_data(t):
                if not v:
                    selects = _gen_select(
                        source_path + escape_name(k),
                        source_tops,
                        {},
                        total_path + escape_name(k),
                        k_total_tops,
                        t,
                    )
                elif is_data(v):
                    selects = _gen_select(
                        source_path + escape_name(k),
                        source_tops,
                        v,
                        total_path + escape_name(k),
                        k_total_tops,
                        t,
                    )
                else:
                    raise Log.error(
                        "Datatype mismatch on {{field}}: Can not merge {{type}} into {{main}}",
                        field=join_field(source_path + escape_name(k)),
                        type=v,
                        main=t,
                    )
                if selects:
                    inner = [
                        ConcatSQL(
                            SQL_SELECT_AS_STRUCT,
                            JoinSQL(ConcatSQL(SQL_COMMA, SQL_CR), selects),
                        )
                    ]
                    selection.append(sql_alias(sql_call("", *inner), escape_name(k)))
            elif is_text(t):
                if is_text(k_tops):
                    # THE SOURCE HAS THIS PROPERTY AS A TOP_LEVEL_FIELD
                    selection.append(
                        ConcatSQL(SQL(k_tops), SQL_AS, quote_column(escape_name(k)))
                    )
                elif v == t:
                    selection.append(
                        ConcatSQL(
                            quote_column(total_path + escape_name(k)),
                            SQL_AS,
                            quote_column(escape_name(k)),
                        )
                    )
                else:
                    if v:
                        Log.note(
                            "Datatype mismatch on {{field}}: Can not merge {{type}} into {{main}}",
                            field=join_field(source_path + escape_name(k)),
                            type=v,
                            main=t,
                        )
                    selection.append(
                        ConcatSQL(
                            sql_call(
                                "CAST",
                                ConcatSQL(
                                    SQL_NULL, SQL_AS, SQL(json_type_to_bq_type[t])
                                ),
                            ),
                            SQL_AS,
                            quote_column(escape_name(k)),
                        )
                    )
            else:
                Log.error("not expected")
        return selection

    output = _gen_select(
        ApiName(),
        flake.top_level_fields,
        flake.schema,
        ApiName(),
        total_flake.top_level_fields,
        total_flake.schema,
    )
    tops = []

    for path, name in jx.sort(total_flake.top_level_fields.leaves(), 0):
        source = flake.top_level_fields[path]
        if source:
            # ALREADY TOP LEVEL FIELD
            source = SQL(source)
        else:
            # PULL OUT TOP LEVEL FIELD
            column = first(flake.leaves(path))
            source = SQL(column.es_column)

        tops.append(ConcatSQL(source, SQL_AS, quote_column(ApiName(name))))

    return tops + output


DEFAULT_SCHEMA = {"_id": INTEGER}
DEFAULT_TYPED_SCHEMA = {"_id": {INTEGER_TYPE: INTEGER}}



# encoding: utf-8
#
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http:# mozilla.org/MPL/2.0/.
#
# Contact: Kyle Lahnakoski (kyle@lahnakoski.com)
#
from __future__ import absolute_import, division, unicode_literals

from jx_base.expressions import RangeOp as RangeOp_
from jx_bigquery.expressions._utils import check
from mo_dots import wrap
from mo_sql import SQL_CASE, SQL_ELSE, SQL_END, SQL_NULL, SQL_THEN, SQL_WHEN


class RangeOp(RangeOp_):
    @check
    def to_bq(self, schema, not_null=False, boolean=False):
        when = self.when.to_bq(schema, boolean=True)[0].sql
        then = self.then.to_bq(schema, not_null=not_null)[0].sql
        els_ = self.els_.to_bq(schema, not_null=not_null)[0].sql
        output = {}
        for t in "bsn":
            if then[t] == None:
                if els_[t] == None:
                    pass
                else:
                    output[t] = (
                        SQL_CASE
                        + SQL_WHEN
                        + when.b
                        + SQL_THEN
                        + SQL_NULL
                        + SQL_ELSE
                        + els_[t]
                        + SQL_END
                    )
            else:
                if els_[t] == None:
                    output[t] = (
                        SQL_CASE + SQL_WHEN + when.b + SQL_THEN + then[t] + SQL_END
                    )
                else:
                    output[t] = (
                        SQL_CASE
                        + SQL_WHEN
                        + when.b
                        + SQL_THEN
                        + then[t]
                        + SQL_ELSE
                        + els_[t]
                        + SQL_END
                    )
        if not output:
            return wrap([{"name": ".", "sql": {"0": SQL_NULL}}])
        else:
            return wrap([{"name": ".", "sql": output}])



# encoding: utf-8
#
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http:# mozilla.org/MPL/2.0/.
#
# Contact: Kyle Lahnakoski (kyle@lahnakoski.com)
#
from __future__ import absolute_import, division, unicode_literals

from jx_base.expressions import RightOp as RightOp_, ZERO, simplified
from jx_bigquery.expressions.basic_substring_op import BasicSubstringOp
from jx_bigquery.expressions.length_op import LengthOp
from jx_bigquery.expressions.max_op import MaxOp
from jx_bigquery.expressions.min_op import MinOp
from jx_bigquery.expressions.sub_op import SubOp


class RightOp(RightOp_):
    @simplified
    def partial_eval(self):
        value = self.value.partial_eval()
        length = self.length.partial_eval()
        max_length = LengthOp(value)

        return BasicSubstringOp(
            [
                value,
                MaxOp([ZERO, MinOp([max_length, SubOp([max_length, length])])]),
                max_length,
            ]
        )



# encoding: utf-8
#
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http:# mozilla.org/MPL/2.0/.
#
# Contact: Kyle Lahnakoski (kyle@lahnakoski.com)
#
from __future__ import absolute_import, division, unicode_literals

from jx_base.expressions import AbsOp as AbsOp_
from jx_bigquery.expressions._utils import BQLang, check
from jx_bigquery.expressions.bql_script import BQLScript
from mo_json import NUMBER


class AbsOp(AbsOp_):
    @check
    def to_bq(self, schema, not_null=False, boolean=False):
        expr = BQLang[self.term].partial_eval().to_bq(schema)[0].sql.n
        return BQLScript(
            expr="ABS(" + expr + ")",
            data_type=NUMBER,
            frum=self,
            miss=self.missing(),
            schema=schema,
        )



# encoding: utf-8
#
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http:# mozilla.org/MPL/2.0/.
#
# Contact: Kyle Lahnakoski (kyle@lahnakoski.com)
#
from __future__ import absolute_import, division, unicode_literals

from jx_base.expressions import BasicStartsWithOp as BasicStartsWithOp_, ONE
from jx_base.expressions.literal import is_literal
from jx_bigquery.expressions._utils import BQLang, check
from jx_bigquery.expressions.length_op import LengthOp
from jx_bigquery.expressions.sql_eq_op import SqlEqOp
from jx_bigquery.expressions.sql_substr_op import SqlSubstrOp
from mo_dots import wrap
from mo_sql import SQL, ConcatSQL, SQL_LIKE, SQL_ESCAPE


class BasicStartsWithOp(BasicStartsWithOp_):
    @check
    def to_bq(self, schema, not_null=False, boolean=False):
        prefix = BQLang[self.prefix].partial_eval()
        if is_literal(prefix):
            value = BQLang[self.value].partial_eval().to_bq(schema)[0].sql.s
            prefix = prefix.to_bq(schema)[0].sql.s
            if "%" in prefix or "_" in prefix:
                for r in "\\_%":
                    prefix = prefix.replaceAll(r, "\\" + r)
                sql = ConcatSQL(value, SQL_LIKE, prefix, SQL_ESCAPE, SQL("\\"))
            else:
                sql = ConcatSQL(value, SQL_LIKE, prefix)
            return wrap([{"name": ".", "sql": {"b": sql}}])
        else:
            return (
                SqlEqOp([SqlSubstrOp([self.value, ONE, LengthOp(prefix)]), prefix])
                .partial_eval()
                .to_bq()
            )



from jx_bigquery.expressions._utils import BQLang
from jx_bigquery.expressions.add_op import AddOp
from jx_bigquery.expressions.and_op import AndOp
from jx_bigquery.expressions.basic_add_op import BasicAddOp
from jx_bigquery.expressions.basic_eq_op import BasicEqOp
from jx_bigquery.expressions.basic_index_of_op import BasicIndexOfOp
from jx_bigquery.expressions.basic_mul_op import BasicMulOp
from jx_bigquery.expressions.basic_starts_with_op import BasicStartsWithOp
from jx_bigquery.expressions.basic_substring_op import BasicSubstringOp
from jx_bigquery.expressions.boolean_op import BooleanOp
from jx_bigquery.expressions.case_op import CaseOp
from jx_bigquery.expressions.coalesce_op import CoalesceOp
from jx_bigquery.expressions.concat_op import ConcatOp
from jx_bigquery.expressions.count_op import CountOp
from jx_bigquery.expressions.date_op import DateOp
from jx_bigquery.expressions.div_op import DivOp
from jx_bigquery.expressions.eq_op import EqOp
from jx_bigquery.expressions.exists_op import ExistsOp
from jx_bigquery.expressions.exp_op import ExpOp
from jx_bigquery.expressions.find_op import FindOp
from jx_bigquery.expressions.first_op import FirstOp
from jx_bigquery.expressions.floor_op import FloorOp
from jx_bigquery.expressions.gt_op import GtOp
from jx_bigquery.expressions.gte_op import GteOp
from jx_bigquery.expressions.in_op import InOp
from jx_bigquery.expressions.integer_op import IntegerOp
from jx_bigquery.expressions.leaves_op import LeavesOp
from jx_bigquery.expressions.length_op import LengthOp
from jx_bigquery.expressions.literal import Literal
from jx_bigquery.expressions.lt_op import LtOp
from jx_bigquery.expressions.lte_op import LteOp
from jx_bigquery.expressions.max_op import MaxOp
from jx_bigquery.expressions.min_op import MinOp
from jx_bigquery.expressions.missing_op import MissingOp
from jx_bigquery.expressions.mul_op import MulOp
from jx_bigquery.expressions.ne_op import NeOp
from jx_bigquery.expressions.not_left_op import NotLeftOp
from jx_bigquery.expressions.not_op import NotOp
from jx_bigquery.expressions.number_op import NumberOp
from jx_bigquery.expressions.or_op import OrOp
from jx_bigquery.expressions.prefix_op import PrefixOp
from jx_bigquery.expressions.string_op import StringOp
from jx_bigquery.expressions.sub_op import SubOp
from jx_bigquery.expressions.suffix_op import SuffixOp
from jx_bigquery.expressions.tuple_op import TupleOp
from jx_bigquery.expressions.variable import Variable
from jx_bigquery.expressions.when_op import WhenOp


BQLang.register_ops(vars())



# encoding: utf-8
#
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http:# mozilla.org/MPL/2.0/.
#
# Contact: Kyle Lahnakoski (kyle@lahnakoski.com)
#
from __future__ import absolute_import, division, unicode_literals

from jx_base.expressions import LeftOp as LeftOp_, ONE
from jx_bigquery.expressions._utils import check
from jx_bigquery.expressions.sql_substr_op import SqlSubstrOp


class LeftOp(LeftOp_):
    @check
    def to_bq(self, schema, not_null=False, boolean=False):
        return SqlSubstrOp([self.value, ONE, self.length]).partial_eval().to_bq(schema)



# encoding: utf-8
#
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http:# mozilla.org/MPL/2.0/.
#
# Contact: Kyle Lahnakoski (kyle@lahnakoski.com)
#
from __future__ import absolute_import, division, unicode_literals

from jx_base.language import Language

from jx_base.expressions import (
    FALSE,
    FalseOp,
    NULL,
    NullOp,
    TrueOp,
    extend,
)
from mo_dots import wrap, FlatList, is_data
from mo_future import decorate
from mo_json import BOOLEAN, NESTED, OBJECT, STRING, NUMBER, IS_NULL
from mo_logs import Log
from mo_sql import (
    SQL,
    SQL_FALSE,
    SQL_NULL,
    SQL_TRUE,
    sql_iso,
    SQL_ZERO,
    SQL_ONE,
    SQL_PLUS,
    SQL_STAR,
    SQL_LT,
)

NumberOp, OrOp, BQLScript = [None] * 3


def check(func):
    """
    TEMPORARY TYPE CHECKING TO ENSURE to_bq() IS OUTPUTTING THE CORRECT FORMAT
    """

    @decorate(func)
    def to_bq(self, schema, not_null=False, boolean=False, **kwargs):
        if kwargs.get("many") != None:
            Log.error("not expecting many")
        try:
            output = func(self, schema, not_null, boolean)
        except Exception as e:
            Log.error("not expected", cause=e)
        if isinstance(output, BQLScript):
            return output
        if not isinstance(output, FlatList):
            Log.error("expecting FlatList")
        if not is_data(output[0].sql):
            Log.error("expecting Data")
        for k, v in output[0].sql.items():
            if k not in {"b", "n", "s", "j", "0"}:
                Log.error("expecting datatypes")
            if not isinstance(v, SQL):
                Log.error("expecting text")
        return output

    return to_bq


@extend(NullOp)
@check
def to_bq(self, schema, not_null=False, boolean=False):
    return wrap([{"name": ".", "sql": {"0": SQL_NULL}}])


@extend(TrueOp)
@check
def to_bq(self, schema, not_null=False, boolean=False):
    return wrap([{"name": ".", "sql": {"b": SQL_TRUE}}])


@extend(FalseOp)
@check
def to_bq(self, schema, not_null=False, boolean=False):
    return wrap([{"name": ".", "sql": {"b": SQL_FALSE}}])


def _inequality_to_bq(self, schema, not_null=False, boolean=False, many=True):
    op, identity = _sql_operators[self.op]
    lhs = NumberOp(self.lhs).partial_eval().to_bq(schema, not_null=True)[0].sql.n
    rhs = NumberOp(self.rhs).partial_eval().to_bq(schema, not_null=True)[0].sql.n
    sql = sql_iso(lhs) + op + sql_iso(rhs)

    output = BQLScript(
        data_type=BOOLEAN,
        expr=sql,
        frum=self,
        miss=OrOp([self.lhs.missing(), self.rhs.missing()]),
        schema=schema,
    )
    return output


@check
def _binaryop_to_bq(self, schema, not_null=False, boolean=False, many=True):
    op, identity = _sql_operators[self.op]

    lhs = NumberOp(self.lhs).partial_eval().to_bq(schema, not_null=True)[0].sql.n
    rhs = NumberOp(self.rhs).partial_eval().to_bq(schema, not_null=True)[0].sql.n
    script = sql_iso(lhs) + op + sql_iso(rhs)
    if not_null:
        sql = script
    else:
        missing = OrOp([self.lhs.missing(), self.rhs.missing()]).partial_eval()
        if missing is FALSE:
            sql = script
        else:
            sql = (
                "CASE WHEN "
                + missing.to_bq(schema, boolean=True)[0].sql.b
                + " THEN NULL ELSE "
                + script
                + " END"
            )
    return wrap([{"name": ".", "sql": {"n": sql}}])


def multiop_to_bq(self, schema, not_null=False, boolean=False, many=False):
    sign, zero = _sql_operators[self.op]
    if len(self.terms) == 0:
        return BQLang[self.default].to_bq(schema)
    elif self.default is NULL:
        return sign.join(
            "COALESCE(" + BQLang[t].to_bq(schema) + ", " + zero + ")"
            for t in self.terms
        )
    else:
        return (
            "COALESCE("
            + sign.join("(" + BQLang[t].to_bq(schema) + ")" for t in self.terms)
            + ", "
            + BQLang[self.default].to_bq(schema)
            + ")"
        )


def basic_multiop_to_bq(self, schema, not_null=False, boolean=False, many=False):
    op, identity = _sql_operators[self.op.split("basic.")[1]]
    sql = op.join(sql_iso(BQLang[t].to_bq(schema)[0].sql.n) for t in self.terms)
    return wrap([{"name": ".", "sql": {"n": sql}}])


BQLang = Language("BQLang")


_sql_operators = {
    # (operator, zero-array default value) PAIR
    "add": (SQL_PLUS, SQL_ZERO),
    "sum": (SQL_PLUS, SQL_ZERO),
    "mul": (SQL_STAR, SQL_ONE),
    "sub": (SQL(" - "), None),
    "div": (SQL(" / "), None),
    "exp": (SQL(" ** "), None),
    "mod": (SQL(" % "), None),
    "gt": (SQL(" > "), None),
    "gte": (SQL(" >= "), None),
    "lte": (SQL(" <= "), None),
    "lt": (SQL_LT, None),
}


json_type_to_bq_type = {
    IS_NULL: "0",
    BOOLEAN: "b",
    NUMBER: "n",
    STRING: "s",
    OBJECT: "j",
    NESTED: "N",
}

sql_type_to_json_type = {
    None: None,
    "0": IS_NULL,
    "b": BOOLEAN,
    "n": NUMBER,
    "s": STRING,
    "j": OBJECT,
}



# encoding: utf-8
#
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http:# mozilla.org/MPL/2.0/.
#
# Contact: Kyle Lahnakoski (kyle@lahnakoski.com)
#
from __future__ import absolute_import, division, unicode_literals

from jx_base.expressions import MissingOp as MissingOp_
from jx_base.language import is_op
from jx_bigquery.expressions._utils import BQLang, check
from mo_dots import wrap
from mo_sql import (
    SQL_AND,
    SQL_EMPTY_STRING,
    SQL_FALSE,
    SQL_IS_NULL,
    SQL_OR,
    SQL_TRUE,
    sql_iso,
)


class MissingOp(MissingOp_):
    @check
    def to_bq(self, schema, not_null=False, boolean=False):
        value = BQLang[self.expr].partial_eval()
        missing_value = value.missing().partial_eval()

        if not is_op(missing_value, MissingOp):
            return missing_value.to_bq(schema)

        value_sql = value.to_bq(schema)

        if len(value_sql) > 1:
            return wrap([{"name": ".", "sql": {"b": SQL_FALSE}}])

        acc = []
        for c in value_sql:
            for t, v in c.sql.items():
                if t == "b":
                    acc.append(sql_iso(v) + SQL_IS_NULL)
                if t == "s":
                    acc.append(
                        sql_iso(sql_iso(v) + SQL_IS_NULL)
                        + SQL_OR
                        + sql_iso(sql_iso(v) + "=" + SQL_EMPTY_STRING)
                    )
                if t == "n":
                    acc.append(sql_iso(v) + SQL_IS_NULL)

        if not acc:
            return wrap([{"name": ".", "sql": {"b": SQL_TRUE}}])
        else:
            return wrap([{"name": ".", "sql": {"b": SQL_AND.join(acc)}}])



# encoding: utf-8
#
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http:# mozilla.org/MPL/2.0/.
#
# Contact: Kyle Lahnakoski (kyle@lahnakoski.com)
#
from __future__ import absolute_import, division, unicode_literals
import string
from enum import EnumMeta

import mo_math
from jx_bigquery.sql import escape_name, TIMESTAMP_FORMAT, unescape_name, ApiName
from jx_python import jx
from mo_dots import is_many, is_data, wrap, split_field, join_field, Data, SLOT, FlatList, NullType, DataObject, \
    set_default
from mo_future import is_text, text, generator_types
from mo_json import (
    BOOLEAN,
    NUMBER,
    STRING,
    OBJECT,
    NESTED,
    python_type_to_json_type,
    INTEGER,
    INTERVAL,
    TIME,
)
from mo_logs import Log
from mo_times.dates import parse

ALLOWED = string.ascii_letters + string.digits


def typed_encode(value, flake):
    """
    RETURN (typed_value, flake_update, added_nested) TUPLES
    :param value: THE RECORD TO CONVERT TO STRICT TYPED FORM
    :param flake: LOOKUP SCHEMA, WILL BE UPDATED WITH CHANGES
    :param top_level_fields: MAP TO TOP LEVEL FIELDS
    :return: (record, update, nested) TUPLE
    """
    _ = flake.columns  # ENSURE WE HAVE INTERNAL STRUCTURES FILLED
    output, update, nested = _typed_encode(value, flake.schema)
    if update:
        # REFRESH COLUMNS
        flake._columns = None
        _ = flake.columns

    worker = wrap(output)
    for path, field in flake._top_level_fields.items():
        worker[field] = worker[path]
        worker[path] = None

        # DO NOT LEAVE ANY EMPTY OBJECT RESIDUE
        _path = split_field(path)
        for i, _ in jx.reverse(enumerate(_path)):
            sub_path = join_field(_path[:i])
            if not worker[sub_path].keys():
                worker[sub_path] = None
            else:
                break

    return output, update, nested


def _typed_encode(value, schema):
    """
    RETURN TRIPLE
    output - THE ENCODED VALUE
    update - THE ADDITIONAL SCHEMA OVER schema PROVIDED
    nested - True IF NESTING IS REQUIRED (CONSIDERED SERIOUS SCHEMA CHANGE)
    """
    if is_many(value):
        if len(value) == 0:
            return None, None, False
        output = []
        update = {}
        nest_added = False
        child_schema = schema.get(NESTED_TYPE)
        if not child_schema:
            nest_added = True
            child_schema = schema[NESTED_TYPE] = {}

        for r in value:
            v, m, n = _typed_encode(r, child_schema)
            output.append(v)
            set_default(update, m)
            nest_added |= n

        if update:
            return {text(REPEATED): output}, {NESTED_TYPE: update}, nest_added
        else:
            return {text(REPEATED): output}, None, nest_added
    elif NESTED_TYPE in schema:
        if not value:
            return {text(REPEATED): []}, None, False
        else:
            return _typed_encode([value], schema)
    elif is_data(value):
        output = {}
        update = {}
        nest_added = False
        for k, v in value.items():
            child_schema = schema.get(k)
            if not child_schema:
                child_schema = schema[k] = {}
            result, more_update, n = _typed_encode(v, child_schema)
            if result != None:
                output[text(escape_name(k))] = result
            set_default(update, {k: more_update})
            nest_added |= n
        return output, update or None, nest_added
    elif is_text(schema):
        v, inserter_type, json_type = schema_type(value)
        if schema != json_type:
            Log.error(
                "Can not convert {{existing_type}} to {{expected_type}}",
                existing_type=json_type,
                expected_type=schema,
            )
        return v, None, False
    elif value is None:
        return {text(escape_name(t)): None for t, child_schema in schema.items()} or None, None, False
    else:
        v, inserter_type, json_type = schema_type(value)
        child_schema = schema.get(inserter_type)
        update = None
        if not child_schema:
            if schema.get(TIME_TYPE):
                # ATTEMPT TO CONVERT TO TIME, IF EXPECTING TIME
                try:
                    v = parse(v).format(TIMESTAMP_FORMAT)
                    return {text(escape_name(TIME_TYPE)): v}, update, False
                except Exception as e:
                    Log.warning(
                        "Failed attempt to convert {{value}} to TIMESTAMP string",
                        value=v,
                        cause=e,
                    )

            schema[inserter_type] = json_type
            update = {inserter_type: json_type}
        return {text(escape_name(inserter_type)): v}, update, False


def schema_type(value):
    clazz = value.__class__
    if clazz.__class__ == EnumMeta:
        return value.name, json_type_to_inserter_type[STRING], STRING

    jt = python_type_to_json_type[clazz]
    if jt == TIME:
        v = parse(value).format(TIMESTAMP_FORMAT)
    elif jt == NUMBER:
        if mo_math.is_finite(value):
            v = float(value)
        else:
            v = None
    else:
        v = value
    return v, json_type_to_inserter_type[jt], jt


def untyped(value):
    """
    REMOVE TYPING AND ESCAPING FROM THE value
    """
    return _untype_value(value)


def _untype_list(value):
    if any(is_data(v) for v in value):
        # MAY BE MORE TYPED OBJECTS IN THIS LIST
        output = [_untype_value(v) for v in value]
    else:
        # LIST OF PRIMITIVE VALUES
        output = value

    if len(output) == 0:
        return None
    elif len(output) == 1:
        return output[0]
    else:
        return output


def _untype_dict(value):
    output = None

    for k, v in value.items():
        k = unescape_name(ApiName(k))
        if k == NESTED_TYPE:
            return _untype_list(v)
        elif k in typed_to_bq_type:
            vv = _untype_value(v)
            if vv is not None:
                # WE ASSUME A PRIMITIVE VALUE, NOT SOME GREATER STRUCTURE
                return vv
        else:
            new_v = _untype_value(v)
            if new_v is not None:
                if output is None:
                    output = {k: new_v}
                else:
                    output[k] = new_v
    return output


_get = object.__getattribute__


def _untype_value(value):
    _type = value.__class__
    if _type is dict:
        return _untype_dict(value)
    elif _type is list:
        return _untype_list(value)
    else:
        return value


json_type_to_bq_type = {
    BOOLEAN: "BOOLEAN",
    NUMBER: "FLOAT64",
    INTEGER: "INT64",
    TIME: "TIMESTAMP",
    INTERVAL: "FLOAT64",
    STRING: "STRING",
    NESTED: "RECORD",
}

bq_type_to_json_type = {
    "INTEGER": INTEGER,
    "INT64": INTEGER,
    "FLOAT64": NUMBER,
    "FLOAT": NUMBER,
    "NUMERIC": NUMBER,
    "BOOL": BOOLEAN,
    "STRING": STRING,
    "BYTES": STRING,
    "DATE": TIME,
    "DATETIME": TIME,
    "TIME": INTERVAL,
    "TIMESTAMP": TIME,
    "RECORD": OBJECT,
}

BOOLEAN_TYPE = "_b_"
INTEGER_TYPE = "_i_"
NUMBER_TYPE = "_n_"
TIME_TYPE = "_t_"
STRING_TYPE = "_s_"
NESTED_TYPE = "_a_"

json_type_to_inserter_type = {
    BOOLEAN: BOOLEAN_TYPE,
    INTEGER: INTEGER_TYPE,
    NUMBER: NUMBER_TYPE,
    TIME: TIME_TYPE,
    INTERVAL: NUMBER_TYPE,
    STRING: STRING_TYPE,
    NESTED: NESTED_TYPE,
}

typed_to_bq_type = {
    BOOLEAN_TYPE: {"field_type": "BOOLEAN", "mode": "NULLABLE"},
    NUMBER_TYPE: {"field_type": "FLOAT64", "mode": "NULLABLE"},
    INTEGER_TYPE: {"field_type": "INT64", "mode": "NULLABLE"},
    TIME_TYPE: {"field_type": "TIMESTAMP", "mode": "NULLABLE"},
    STRING_TYPE: {"field_type": "STRING", "mode": "NULLABLE"},
    NESTED_TYPE: {"field_type": "RECORD", "mode": "REPEATED"},
}

REPEATED = escape_name(NESTED_TYPE)



# encoding: utf-8
#
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http:# mozilla.org/MPL/2.0/.
#
# Contact: Kyle Lahnakoski (kyle@lahnakoski.com)
#
from __future__ import absolute_import, division, unicode_literals

from jx_base.expressions import AndOp as AndOp_
from jx_bigquery.expressions._utils import BQLang, check
from mo_dots import wrap
from mo_sql import SQL_AND, SQL_FALSE, SQL_TRUE, sql_iso


class AndOp(AndOp_):
    @check
    def to_bq(self, schema, not_null=False, boolean=False):
        if not self.terms:
            return wrap([{"name": ".", "sql": {"b": SQL_TRUE}}])
        elif all(self.terms):
            return wrap(
                [
                    {
                        "name": ".",
                        "sql": {
                            "b": SQL_AND.join(
                                [
                                    sql_iso(
                                        BQLang[t].to_bq(schema, boolean=True)[0].sql.b
                                    )
                                    for t in self.terms
                                ]
                            )
                        },
                    }
                ]
            )
        else:
            return wrap([{"name": ".", "sql": {"b": SQL_FALSE}}])



# encoding: utf-8
#
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http:# mozilla.org/MPL/2.0/.
#
# Contact: Kyle Lahnakoski (kyle@lahnakoski.com)
#
from __future__ import absolute_import, division, unicode_literals

from copy import copy

from google.cloud.bigquery import SchemaField

import jx_base
from jx_bigquery.partitions import Partition
from jx_bigquery.sql import unescape_name, ApiName, escape_name
from jx_bigquery.typed_encoder import (
    bq_type_to_json_type,
    NESTED_TYPE,
    typed_to_bq_type,
    REPEATED,
    json_type_to_bq_type,
    json_type_to_inserter_type,
)
from jx_python import jx
from mo_dots import join_field, startswith_field, coalesce, Data, wrap, split_field, Null
from mo_future import is_text, first, sort_using_key, text, OrderedDict
from mo_json import NESTED, STRUCT, OBJECT, STRING, NUMBER, scrub
from mo_logs import Log, Except
from mo_times.dates import Date


DEBUG = True


class Snowflake(jx_base.Snowflake):
    """
    REPRESENTS A JSON SCHEMA, PLUS SOME ADDITIONAL INFO TO MAP IT TO THE BQ SCHEMA
    THE TWO BIGGEST COMPLICATIONS ARE
    1. A FEW PROPERTIES MUST BE MAPPED TO A TOP-LEVEL, WHICH COMPLICATES ALL THE TRANSLATION LOGIC
    2. THE PARTITION FIELD MUST BE A TIMESTAMP, WHICH IS NOT A JSON TYPE
    """

    def __init__(self, es_index, top_level_fields, partition, schema=None):
        """
        :param es_index:  NAME OF THE INDEX (FOR PROVIDING FULL COLUMN DETAILS)
        :param top_level_fields:  REQUIRED TO MAP INNER PROPERTIES TO THE TOP LEVEL, AS REQUIRED BY BQ FOR PARTITIONS AND CLUSTERING
        :param partition:  THE partition.field MUST BE KNOWN SO IT CAN BE CONVERTED FROM UNIX TIME TO bq TIMESTAMP
        """
        self.schema = schema or {}
        if DEBUG:
            if not is_text(es_index):
                Log.error("expecting string")
            if any(len(split_field(k)) > 1 for k in self.schema.keys()):
                Log.error("expecting schema to be a deep structure")
            if any(len(split_field(k)) > 1 for k in top_level_fields.keys()):
                Log.error("expecting top level fields to be a deep structure")

        self.es_index = es_index
        self._columns = None
        self.top_level_fields = top_level_fields
        self._top_level_fields = None  # dict FROM FULL-API-NAME TO TOP-LEVEL-FIELD NAME
        self._es_type_info = Data()
        if partition.field:
            self._es_type_info[partition.field] = "TIMESTAMP"
        self.partition = partition
        self._partition = None

    @property
    def bq_time_partitioning(self):
        _ = self.columns  # ENSURE self_partition HAS BEEN MADE
        return self._partition.bq_time_partitioning

    @property
    def columns(self):
        if not self._columns:
            now = Date.now()
            columns = []

            def parse_schema(schema, tops, es_type_info, jx_path, nested_path, es_path):
                if is_text(schema):
                    json_type = schema
                    expected_es_type = json_type_to_bq_type[json_type]
                    if es_type_info and es_type_info != expected_es_type:
                        Log.error(
                            "expecting {{path}} to be of type {{expected_type}} not of type {{observed_type}}",
                            path=jx_path,
                            expected_type=expected_es_type,
                            observed_type=es_type_info,
                        )
                    c = jx_base.Column(
                        name=join_field(jx_path),
                        es_column=coalesce(tops, text(es_path)),
                        es_index=self.es_index,
                        es_type=coalesce(es_type_info, expected_es_type),
                        jx_type=json_type,
                        nested_path=nested_path,
                        last_updated=now,
                    )
                    columns.append(c)
                else:
                    c = jx_base.Column(
                        name=join_field(jx_path),
                        es_column=text(es_path),
                        es_index=self.es_index,
                        es_type="RECORD",
                        jx_type=OBJECT,
                        nested_path=nested_path,
                        last_updated=now,
                    )
                    columns.append(c)
                    count = len(columns)
                    for k, s in schema.items():
                        if k == NESTED_TYPE:
                            c.jx_type = NESTED
                            parse_schema(
                                s,
                                tops if is_text(tops) else tops[k],
                                es_type_info
                                if is_text(es_type_info)
                                else es_type_info[k],
                                jx_path + (k,),
                                (jx_path,) + nested_path,
                                es_path + escape_name(k),
                            )
                        else:
                            parse_schema(
                                s,
                                tops if is_text(tops) else tops[k],
                                es_type_info
                                if is_text(es_type_info)
                                else es_type_info[k],
                                jx_path + (k,),
                                nested_path,
                                es_path + escape_name(k),
                            )
                    if is_text(tops) and len(columns) > count + 1:
                        Log.error(
                            "too many top level fields at {{field}}:",
                            field=join_field(jx_path),
                        )

            parse_schema(
                self.schema,
                self.top_level_fields,
                self._es_type_info,
                (),
                (".",),
                ApiName(),
            )
            self._columns = columns

            self._top_level_fields = OrderedDict()  # FORCE ORDERING
            for path, field in jx.sort(wrap(self.top_level_fields).leaves(), 0):
                leaves = self.leaves(path)
                if not leaves:
                    continue
                if len(leaves) > 1:
                    Log.error(
                        "expecting {{path}} to have just one primitive value", path=path
                    )
                specific_path = first(leaves).name
                self._top_level_fields[
                    ".".join(text(escape_name(step)) for step in split_field(specific_path))
                ] = field
            self._partition = Partition(kwargs=self.partition, flake=self)

        return self._columns

    @classmethod
    def parse(cls, big_query_schema, es_index, top_level_fields, partition):
        """
        PARSE A BIGQUERY SCHEMA
        :param schema:  BQ SCHEMA (WHICH IS A list OF SCHEMA OBJECTS
        :param es_index:  NAME OF THE BQ TABLE
        :param top_level_fields: MAP FROM PROPERTY PATH TO TOP-LEVEL-FIELDNAME
        :param partition: SO WE KNOW WHICH FIELD MUST BE A TIMESTAMP
        :return:
        """

        def parse_schema(big_query_schema, jx_path, nested_path, es_path):
            output = OrderedDict()

            if any(ApiName(e.name) == REPEATED for e in big_query_schema):
                big_query_schema = [
                    e for e in big_query_schema if ApiName(e.name) == REPEATED
                ]

            for e in big_query_schema:
                json_type = bq_type_to_json_type[e.field_type]
                name = unescape_name(ApiName(e.name))
                full_name = jx_path + (name,)
                full_es_path = es_path + (e.name,)

                if e.field_type == "RECORD":
                    if e.mode == "REPEATED":
                        output[name] = parse_schema(
                            e.fields, full_name, (jx_path,) + nested_path, full_es_path
                        )
                    else:
                        output[name] = parse_schema(
                            e.fields, full_name, nested_path, full_es_path
                        )
                else:
                    if e.mode == "REPEATED":
                        output[name] = {NESTED_TYPE: json_type}
                    else:
                        output[name] = json_type
            return output

        output = Snowflake(es_index, Data(), partition)

        # GRAB THE TOP-LEVEL FIELDS
        top_fields = [field for path, field in top_level_fields.leaves()]
        i = 0
        while i < len(big_query_schema) and big_query_schema[i].name in top_fields:
            i = i + 1

        output.top_level_fields = top_level_fields
        output.schema = parse_schema(big_query_schema[i:], (), (".",), ())

        # INSERT TOP-LEVEL FIELDS INTO THE loopkup
        schema = wrap(output.schema)
        for column in big_query_schema[:i]:
            path = first(
                name
                for name, field in top_level_fields.leaves()
                if field == column.name
            )
            json_type = bq_type_to_json_type[column.field_type]
            schema[path] = OrderedDict(
                [(json_type_to_inserter_type[json_type], json_type)]
            )
        return output

    def leaves(self, name):
        return [
            c
            for c in self.columns
            if c.jx_type not in STRUCT and startswith_field(c.name, name)
        ]

    def __eq__(self, other):
        if not isinstance(other, Snowflake):
            return False

        def identical_schema(a, b):
            if is_text(b):
                return a == b
            if len(a.keys()) != len(b.keys()):
                return False
            for (ka, va), (kb, vb) in zip(a.items(), b.items()):
                # WARNING!  ASSUMES dicts ARE OrderedDict
                if ka != kb or not identical_schema(va, vb):
                    return False
            return True

        return identical_schema(self.schema, other.schema)

    def __or__(self, other):
        return merge(self, other)

    def to_bq_schema(self):
        top_fields = []

        def _schema_to_bq_schema(jx_path, es_path, schema):
            output = []
            nt = schema.get(NESTED_TYPE)
            if nt:
                schema = {NESTED_TYPE: nt}
            for t, sub_schema in jx.sort(schema.items(), 0):
                bqt = typed_to_bq_type.get(
                    t, {"field_type": "RECORD", "mode": "NULLABLE"}
                )
                full_name = es_path + escape_name(t)
                top_field = self._top_level_fields.get(text(full_name))
                if is_text(sub_schema):
                    new_field_type = json_type_to_bq_type.get(sub_schema, sub_schema)
                    if new_field_type != bqt["field_type"]:
                        # OVERRIDE TYPE
                        bqt = bqt.copy()
                        bqt["field_type"] = new_field_type
                    fields = ()
                else:
                    fields = _schema_to_bq_schema(jx_path + (t,), full_name, sub_schema)

                if top_field:
                    if fields:
                        Log.error("not expecting a structure")
                    if self._partition.field == top_field:
                        if bqt["field_type"] != "TIMESTAMP":
                            Log.error("Partition field must be of time type")
                    struct = SchemaField(name=top_field, fields=fields, **bqt)
                    top_fields.append(struct)
                elif not fields and bqt["field_type"] == "RECORD":
                    # THIS CAN HAPPEN WHEN WE MOVE A PRIMITIVE FIELD TO top_fields
                    pass
                else:
                    struct = SchemaField(
                        name=text(escape_name(t)), fields=fields, **bqt
                    )
                    output.append(struct)
            return output

        _ = self.columns  # ENSURE schema HAS BEEN PROCESSED
        if not self.schema:
            return []
        main_schema = _schema_to_bq_schema((), ApiName(), self.schema)
        output = sort_using_key(top_fields, key=lambda v: v.name) + main_schema
        return output


def merge(schemas, es_index, top_level_fields, partition):
    def _merge(*schemas):
        if len(schemas) == 1:
            return schemas[0]
        try:
            if any(NESTED_TYPE in s for s in schemas):
                # IF THERE ARE ANY ARRAYS, THEN THE MERGE IS AN ARRAY
                new_schemas = []
                for schema in schemas:
                    if NESTED_TYPE in schema:
                        sub_schema = schema[NESTED_TYPE]
                        residue = {k: v for k, v in schema.items() if k != NESTED_TYPE}
                        new_schemas.append(_merge(sub_schema, residue))
                    else:
                        new_schemas.append(schema)
                return {NESTED_TYPE: _merge(*new_schemas)}
            else:
                return OrderedDict(
                    (k, _merge(*(ss for s in schemas for ss in [s.get(k)] if ss)))
                    for k in jx.sort(set(k for s in schemas for k in s.keys()))
                )
        except Exception as e:
            e = Except.wrap(e)
            if "Expecting types to match" in e:
                raise e
            t = list(set(schemas))
            if len(t) == 1:
                return t[0]
            elif len(t) == 2 and STRING in t and NUMBER in t:
                return STRING
            else:
                Log.error("Expecting types to match {{types|json}}", types=t)

    output = Snowflake(
        es_index=es_index, top_level_fields=top_level_fields, partition=partition
    )
    output.schema = _merge(*(s.schema for s in schemas))
    return output



# encoding: utf-8
#
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http:# mozilla.org/MPL/2.0/.
#
# Contact: Kyle Lahnakoski (kyle@lahnakoski.com)
#
from __future__ import absolute_import, division, unicode_literals

from jx_bigquery.sql import quote_list

from jx_base.expressions import InOp as InOp_
from jx_base.language import is_op
from jx_bigquery.expressions._utils import BQLang, check
from jx_bigquery.expressions.literal import Literal
from mo_dots import wrap
from mo_json import json2value
from mo_logs import Log
from mo_sql import SQL_FALSE, SQL_OR, sql_iso, ConcatSQL, SQL_IN


class InOp(InOp_):
    @check
    def to_bq(self, schema, not_null=False, boolean=False):
        if not is_op(self.superset, Literal):
            Log.error("Not supported")
        j_value = json2value(self.superset.json)
        if j_value:
            var = BQLang[self.value].to_bq(schema)
            sql = SQL_OR.join(
                sql_iso(ConcatSQL(v, SQL_IN, quote_list(j_value)))
                for t, v in var[0].sql.items()
            )
        else:
            sql = SQL_FALSE
        return wrap([{"name": ".", "sql": {"b": sql}}])



# encoding: utf-8
#
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http:# mozilla.org/MPL/2.0/.
#
# Contact: Kyle Lahnakoski (kyle@lahnakoski.com)
#
from __future__ import absolute_import, division, unicode_literals

from jx_base.expressions import LengthOp as LengthOp_
from jx_base.expressions.literal import is_literal
from jx_bigquery.expressions._utils import BQLang, check
from jx_bigquery.sql import quote_value
from mo_dots import Null, wrap
from mo_future import text
from mo_json import value2json
from mo_sql import SQL, sql_iso, ConcatSQL


class LengthOp(LengthOp_):
    @check
    def to_bq(self, schema, not_null=False, boolean=False):
        term = BQLang[self.term].partial_eval()
        if is_literal(term):
            val = term.value
            if isinstance(val, text):
                sql = quote_value(len(val))
            elif isinstance(val, (float, int)):
                sql = quote_value(len(value2json(val)))
            else:
                return Null
        else:
            value = term.to_bq(schema, not_null=not_null)[0].sql.s
            sql = ConcatSQL(SQL("LENGTH"), sql_iso(value))
        return wrap([{"name": ".", "sql": {"n": sql}}])



# encoding: utf-8
#
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http:# mozilla.org/MPL/2.0/.
#
# Contact: Kyle Lahnakoski (kyle@lahnakoski.com)
#
from __future__ import absolute_import, division, unicode_literals

from jx_base.expressions import BasicIndexOfOp as BasicIndexOfOp_
from jx_bigquery.expressions._utils import check
from jx_bigquery.expressions.literal import Literal
from mo_dots import wrap
from mo_sql import SQL_CASE, SQL_ELSE, SQL_END, SQL_THEN, SQL_WHEN, sql_iso


class BasicIndexOfOp(BasicIndexOfOp_):
    @check
    def to_bq(self, schema, not_null=False, boolean=False):
        value = self.value.to_bq(schema)[0].sql.s
        find = self.find.to_bq(schema)[0].sql.s
        start = self.start

        if isinstance(start, Literal) and start.value == 0:
            return wrap(
                [
                    {
                        "name": ".",
                        "sql": {"n": "INSTR" + sql_iso(value + "," + find) + "-1"},
                    }
                ]
            )
        else:
            start_index = start.to_bq(schema)[0].sql.n
            found = "INSTR(SUBSTR" + sql_iso(value + "," + start_index + "+1)," + find)
            return wrap(
                [
                    {
                        "name": ".",
                        "sql": {
                            "n": (
                                SQL_CASE
                                + SQL_WHEN
                                + found
                                + SQL_THEN
                                + found
                                + "+"
                                + start_index
                                + "-1"
                                + SQL_ELSE
                                + "-1"
                                + SQL_END
                            )
                        },
                    }
                ]
            )



# encoding: utf-8
#
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http:# mozilla.org/MPL/2.0/.
#
# Contact: Kyle Lahnakoski (kyle@lahnakoski.com)
#
from __future__ import absolute_import, division, unicode_literals

from jx_base.expressions import NotOp as NotOp_
from jx_base.language import is_op
from jx_bigquery.expressions._utils import check
from jx_bigquery.expressions.boolean_op import BooleanOp
from mo_dots import wrap
from mo_sql import sql_iso


class NotOp(NotOp_):
    @check
    def to_bq(self, schema, not_null=False, boolean=False):
        not_expr = NotOp(BooleanOp(self.term)).partial_eval()
        if is_op(not_expr, NotOp):
            return wrap(
                [
                    {
                        "name": ".",
                        "sql": {
                            "b": "NOT " + sql_iso(not_expr.term.to_bq(schema)[0].sql.b)
                        },
                    }
                ]
            )
        else:
            return not_expr.to_bq(schema)



# encoding: utf-8
#
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http:# mozilla.org/MPL/2.0/.
#
# Contact: Kyle Lahnakoski (kyle@lahnakoski.com)
#
from __future__ import absolute_import, division, unicode_literals

from jx_base.expressions import FALSE, SqlEqOp as SqlEqOp_
from jx_base.expressions.literal import is_literal
from jx_bigquery.expressions._utils import BQLang, check
from jx_bigquery.expressions.boolean_op import BooleanOp
from mo_dots import wrap
from mo_logs import Log
from mo_sql import SQL_IS_NULL, SQL_OR, sql_iso, ConcatSQL, JoinSQL, SQL_EQ


class SqlEqOp(SqlEqOp_):
    @check
    def to_bq(self, schema, not_null=False, boolean=False):
        lhs = BQLang[self.lhs].partial_eval()
        rhs = BQLang[self.rhs].partial_eval()
        lhs_sql = lhs.to_bq(schema, not_null=True)
        rhs_sql = rhs.to_bq(schema, not_null=True)
        if is_literal(rhs) and lhs_sql[0].sql.b != None and rhs.value in ("T", "F"):
            rhs_sql = BooleanOp(rhs).to_bq(schema)
        if is_literal(lhs) and rhs_sql[0].sql.b != None and lhs.value in ("T", "F"):
            lhs_sql = BooleanOp(lhs).to_bq(schema)

        if len(lhs_sql) != len(rhs_sql):
            Log.error("lhs and rhs have different dimensionality!?")

        acc = []
        for l, r in zip(lhs_sql, rhs_sql):
            for t in "bsnj":
                if r.sql[t] == None:
                    if l.sql[t] == None:
                        pass
                    else:
                        acc.append(ConcatSQL(l.sql[t], SQL_IS_NULL))
                elif l.sql[t] == None:
                    acc.append(ConcatSQL(r.sql[t], SQL_IS_NULL))
                else:
                    acc.append(ConcatSQL(sql_iso(l.sql[t]), SQL_EQ, sql_iso(r.sql[t])))
        if not acc:
            return FALSE.to_bq(schema)
        else:
            return wrap([{"name": ".", "sql": {"b": JoinSQL(SQL_OR, acc)}}])



# encoding: utf-8
#
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http:# mozilla.org/MPL/2.0/.
#
# Contact: Kyle Lahnakoski (kyle@lahnakoski.com)
#
from __future__ import absolute_import, division, unicode_literals

from jx_base.expressions import FloorOp as FloorOp_
from jx_bigquery.expressions._utils import BQLang, check
from mo_dots import wrap
from mo_future import text
from mo_sql import sql_iso


class FloorOp(FloorOp_):
    @check
    def to_bq(self, schema, not_null=False, boolean=False):
        lhs = BQLang[self.lhs].to_bq(schema)[0].sql.n
        rhs = BQLang[self.rhs].to_bq(schema)[0].sql.n
        modifier = lhs + " < 0 "

        if text(rhs).strip() != "1":
            floor = "CAST" + sql_iso(lhs + "/" + rhs + " AS INTEGER")
            sql = sql_iso(sql_iso(floor) + "-" + sql_iso(modifier)) + "*" + rhs
        else:
            floor = "CAST" + sql_iso(lhs + " AS INTEGER")
            sql = sql_iso(floor) + "-" + sql_iso(modifier)

        return wrap([{"name": ".", "sql": {"n": sql}}])



# encoding: utf-8
#
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http:# mozilla.org/MPL/2.0/.
#
# Contact: Kyle Lahnakoski (kyle@lahnakoski.com)
#
from __future__ import absolute_import, division, unicode_literals

from jx_base.expressions import CaseOp as CaseOp_
from jx_bigquery.expressions._utils import BQLang, check
from mo_dots import coalesce, wrap
from mo_sql import (
    SQL_CASE,
    SQL_ELSE,
    SQL_END,
    SQL_NULL,
    SQL_THEN,
    SQL_WHEN,
    ConcatSQL,
)


class CaseOp(CaseOp_):
    @check
    def to_bq(self, schema, not_null=False, boolean=False):
        if len(self.whens) == 1:
            return BQLang[self.whens[-1]].to_bq(schema)

        output = {}
        for t in "bsn":  # EXPENSIVE LOOP to_bq() RUN 3 TIMES
            els_ = coalesce(BQLang[self.whens[-1]].to_bq(schema)[0].sql[t], SQL_NULL)
            acc = SQL_ELSE + els_ + SQL_END
            for w in reversed(self.whens[0:-1]):
                acc = ConcatSQL(
                    SQL_WHEN,
                    BQLang[w.when].to_bq(schema, boolean=True)[0].sql.b,
                    SQL_THEN,
                    coalesce(BQLang[w.then].to_bq(schema)[0].sql[t], SQL_NULL),
                    acc,
                )
            output[t] = SQL_CASE + acc
        return wrap([{"name": ".", "sql": output}])



# encoding: utf-8
#
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http:# mozilla.org/MPL/2.0/.
#
# Contact: Kyle Lahnakoski (kyle@lahnakoski.com)
#
from __future__ import absolute_import, division, unicode_literals

from jx_bigquery.sql import quote_value

from jx_base.expressions import StringOp as StringOp_
from jx_bigquery.expressions._utils import BQLang, check
from mo_dots import wrap
from mo_sql import (
    SQL_CASE,
    SQL_ELSE,
    SQL_END,
    SQL_NULL,
    SQL_THEN,
    SQL_WHEN,
    sql_coalesce,
    sql_iso,
)


class StringOp(StringOp_):
    @check
    def to_bq(self, schema, not_null=False, boolean=False):
        test = BQLang[self.term].missing().to_bq(schema, boolean=True)[0].sql.b
        value = BQLang[self.term].to_bq(schema, not_null=True)[0].sql
        acc = []
        for t, v in value.items():
            if t == "b":
                acc.append(
                    SQL_CASE
                    + SQL_WHEN
                    + sql_iso(test)
                    + SQL_THEN
                    + SQL_NULL
                    + SQL_WHEN
                    + sql_iso(v)
                    + SQL_THEN
                    + "'true'"
                    + SQL_ELSE
                    + "'false'"
                    + SQL_END
                )
            elif t == "s":
                acc.append(v)
            else:
                acc.append(
                    "RTRIM(RTRIM(CAST"
                    + sql_iso(v + " as TEXT), " + quote_value("0"))
                    + ", "
                    + quote_value(".")
                    + ")"
                )
        if not acc:
            return wrap([{}])
        elif len(acc) == 1:
            return wrap([{"name": ".", "sql": {"s": acc[0]}}])
        else:
            return wrap([{"name": ".", "sql": {"s": sql_coalesce(acc)}}])



# encoding: utf-8
#
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http:# mozilla.org/MPL/2.0/.
#
# Contact: Kyle Lahnakoski (kyle@lahnakoski.com)
#
from __future__ import absolute_import, division, unicode_literals

from jx_base.expressions import CoalesceOp as CoalesceOp_
from jx_bigquery.expressions._utils import BQLang, check
from mo_dots import wrap
from mo_sql import sql_coalesce


class CoalesceOp(CoalesceOp_):
    @check
    def to_bq(self, schema, not_null=False, boolean=False):
        acc = {"b": [], "s": [], "n": []}

        for term in self.terms:
            for t, v in BQLang[term].to_bq(schema)[0].sql.items():
                acc[t].append(v)

        output = {}
        for t, terms in acc.items():
            if not terms:
                continue
            elif len(terms) == 1:
                output[t] = terms[0]
            else:
                output[t] = sql_coalesce(terms)
        return wrap([{"name": ".", "sql": output}])



# encoding: utf-8
#
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http:# mozilla.org/MPL/2.0/.
#
# Contact: Kyle Lahnakoski (kyle@lahnakoski.com)
#
from __future__ import absolute_import, division, unicode_literals

from jx_base.expressions import SqlInstrOp as SqlInstrOp_
from jx_bigquery.expressions._utils import check
from mo_dots import wrap
from mo_sql import sql_iso, sql_list


class SqlInstrOp(SqlInstrOp_):
    @check
    def to_bq(self, schema, not_null=False, boolean=False):
        value = self.value.to_bq(schema, not_null=True)[0].sql.s
        find = self.find.to_bq(schema, not_null=True)[0].sql.s

        return wrap(
            [{"name": ".", "sql": {"n": "INSTR" + sql_iso(sql_list([value, find]))}}]
        )

    def partial_eval(self):
        value = self.value.partial_eval()
        find = self.find.partial_eval()
        return SqlInstrOp([value, find])



# encoding: utf-8
#
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http:# mozilla.org/MPL/2.0/.
#
# Contact: Kyle Lahnakoski (kyle@lahnakoski.com)
#
from __future__ import absolute_import, division, unicode_literals

from jx_base.expressions import CountOp as CountOp_
from jx_bigquery.expressions._utils import BQLang, check
from mo_dots import wrap
from mo_sql import (
    SQL,
    SQL_CASE,
    SQL_ELSE,
    SQL_END,
    SQL_IS_NULL,
    SQL_THEN,
    SQL_TRUE,
    SQL_WHEN,
    sql_iso,
    SQL_ZERO,
    SQL_ONE,
    ConcatSQL,
)


class CountOp(CountOp_):
    @check
    def to_bq(self, schema, not_null=False, boolean=False):
        acc = []
        for term in self.terms:
            sqls = BQLang[term].to_bq(schema)
            if len(sqls) > 1:
                acc.append(SQL_TRUE)
            else:
                for t, v in sqls[0].sql.items():
                    if t in ["b", "s", "n"]:
                        acc.append(
                            ConcatSQL(
                                SQL_CASE,
                                SQL_WHEN,
                                sql_iso(v),
                                SQL_IS_NULL,
                                SQL_THEN,
                                SQL_ZERO,
                                SQL_ELSE,
                                SQL_ONE,
                                SQL_END,
                            )
                        )
                    else:
                        acc.append(SQL_TRUE)

        if not acc:
            return wrap([{}])
        else:
            return wrap([{"nanme": ".", "sql": {"n": SQL("+").join(acc)}}])



# encoding: utf-8
#
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http:# mozilla.org/MPL/2.0/.
#
# Contact: Kyle Lahnakoski (kyle@lahnakoski.com)
#
from __future__ import absolute_import, division, unicode_literals

from jx_base.expressions import FALSE, NULL, ONE, SQLScript as SQLScript_, TRUE, ZERO
from jx_bigquery.expressions._utils import json_type_to_sql_type, SQLang, check
from mo_dots import coalesce, wrap
from mo_future import PY2, text
from mo_logs import Log
from mo_sql import (
    SQL,
    SQL_CASE,
    SQL_END,
    SQL_NULL,
    SQL_THEN,
    SQL_WHEN,
    sql_iso,
    ConcatSQL,
    SQL_NOT,
)


class SQLScript(SQLScript_, SQL):
    __slots__ = ("miss", "data_type", "expr", "frum", "many", "schema")

    def __init__(self, data_type, expr, frum, miss=None, many=False, schema=None):
        object.__init__(self)
        if miss not in [None, NULL, FALSE, TRUE, ONE, ZERO]:
            if frum.lang != miss.lang:
                Log.error("logic error")

        self.miss = coalesce(
            miss, FALSE
        )  # Expression that will return true/false to indicate missing result
        self.data_type = data_type  # JSON DATA TYPE
        self.expr = expr
        self.many = many  # True if script returns multi-value
        self.frum = frum  # THE ORIGINAL EXPRESSION THAT MADE expr
        self.schema = schema

    @property
    def type(self):
        return self.data_type

    @property
    def name(self):
        return "."

    def __getitem__(self, item):
        if not self.many:
            if item == 0:
                return self
            else:
                Log.error("this is a primitive value")
        else:
            Log.error("do not know how to handle")

    def __iter__(self):
        """
        ASSUMED PART OF class SQL, RETURN SQL
        """
        for e in self.expr:
            yield e

    @property
    def sql(self):
        self.miss = self.miss.partial_eval()
        if self.miss is TRUE:
            return wrap({json_type_to_sql_type[self.data_type]: SQL_NULL})
        elif self.miss is FALSE:
            return wrap({json_type_to_sql_type[self.data_type]: self.expr})
        else:
            return wrap(
                {
                    json_type_to_sql_type[self.data_type]: ConcatSQL(
                        SQL_CASE,
                        SQL_WHEN,
                        SQL_NOT,
                        sql_iso(SQLang[self.miss].to_sql(self.schema)[0].sql.b),
                        SQL_THEN,
                        self.expr,
                        SQL_END,
                    )
                }
            )

    def __str__(self):
        return str(self.sql)

    def __unicode__(self):
        return text(self.sql)

    def __add__(self, other):
        return text(self) + text(other)

    def __radd__(self, other):
        return text(other) + text(self)

    if PY2:
        __unicode__ = __str__

    @check
    def to_sql(self, schema, not_null=False, boolean=False, many=True):
        return self

    def missing(self):
        return self.miss

    def __data__(self):
        return {"script": self.script}

    def __eq__(self, other):
        if not isinstance(other, SQLScript_):
            return False
        elif self.expr == other.expr:
            return True
        else:
            return False



# encoding: utf-8
#
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http:# mozilla.org/MPL/2.0/.
#
# Contact: Kyle Lahnakoski (kyle@lahnakoski.com)
#
from __future__ import absolute_import, division, unicode_literals

from jx_base.expressions import SuffixOp as SuffixOp_
from jx_bigquery.expressions._utils import check
from jx_bigquery.expressions.eq_op import EqOp
from jx_bigquery.expressions.length_op import LengthOp
from jx_bigquery.expressions.literal import Literal
from jx_bigquery.expressions.right_op import RightOp
from mo_dots import wrap
from mo_sql import SQL_FALSE, SQL_TRUE


class SuffixOp(SuffixOp_):
    @check
    def to_bq(self, schema, not_null=False, boolean=False):
        if not self.expr:
            return wrap([{"name": ".", "sql": {"b": SQL_FALSE}}])
        elif isinstance(self.suffix, Literal) and not self.suffix.value:
            return wrap([{"name": ".", "sql": {"b": SQL_TRUE}}])
        else:
            return (
                EqOp([RightOp([self.expr, LengthOp(self.suffix)]), self.suffix])
                .partial_eval()
                .to_bq(schema)
            )



# encoding: utf-8
#
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http:# mozilla.org/MPL/2.0/.
#
# Contact: Kyle Lahnakoski (kyle@lahnakoski.com)
#
from __future__ import absolute_import, division, unicode_literals

from jx_base.expressions import NULL, SqlSubstrOp as SqlSubstrOp_
from jx_bigquery.expressions._utils import check
from jx_bigquery.expressions.literal import Literal
from mo_dots import wrap
from mo_sql import sql_iso, sql_list


class SqlSubstrOp(SqlSubstrOp_):
    @check
    def to_bq(self, schema, not_null=False, boolean=False):
        value = self.value.to_bq(schema, not_null=True)[0].sql.s
        start = self.start.to_bq(schema, not_null=True)[0].sql.n
        if self.length is NULL:
            sql = "SUBSTR" + sql_iso(sql_list([value, start]))
        else:
            length = self.length.to_bq(schema, not_null=True)[0].sql.n
            sql = "SUBSTR" + sql_iso(sql_list([value, start, length]))
        return wrap([{"name": ".", "sql": sql}])

    def partial_eval(self):
        value = self.value.partial_eval()
        start = self.start.partial_eval()
        length = self.length.partial_eval()
        if isinstance(start, Literal) and start.value == 1:
            if length is NULL:
                return value
        return SqlSubstrOp([value, start, length])



# encoding: utf-8
#
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http:# mozilla.org/MPL/2.0/.
#
# Contact: Kyle Lahnakoski (kyle@lahnakoski.com)
#
from __future__ import absolute_import, division, unicode_literals

from jx_base.expressions import NotLeftOp as NotLeftOp_
from jx_bigquery.expressions._utils import check
from mo_dots import wrap


class NotLeftOp(NotLeftOp_):
    @check
    def to_bq(self, schema, not_null=False, boolean=False):
        # test_v = self.value.missing().to_bq(boolean=True)[0].sql.b
        # test_l = self.length.missing().to_bq(boolean=True)[0].sql.b
        v = self.value.to_bq(schema, not_null=True)[0].sql.s
        l = "max(0, " + self.length.to_bq(schema, not_null=True)[0].sql.n + ")"

        expr = "substr(" + v + ", " + l + "+1)"
        return wrap([{"name": ".", "sql": {"s": expr}}])



# encoding: utf-8
# THIS FILE IS AUTOGENERATED!
from __future__ import unicode_literals
from setuptools import setup
setup(
    author='Kyle Lahnakoski',
    author_email='kyle@lahnakoski.com',
    classifiers=["Development Status :: 4 - Beta","Topic :: Software Development :: Libraries","Topic :: Software Development :: Libraries :: Python Modules","License :: OSI Approved :: Mozilla Public License 2.0 (MPL 2.0)","Programming Language :: Python :: 3.6","Programming Language :: Python :: 3.7","Programming Language :: Python :: 3.8"],
    description='jx-bigquery - JSON Expressions for BigQuery',
    include_package_data=True,
    install_requires=["google-cloud-bigquery","jx-python>=3.57.20089","jx-python>=3.57.20089","mo-dots>=3.53.20059","mo-future>=3.51.20059","mo-json>=3.57.20089","mo-kwargs>=3.54.20059","mo-logs>=3.57.20089","mo-sql>=3.55.20074","mo-times>=3.57.20089","mo-times>=3.57.20089"],
    license='MPL 2.0',
    long_description='# jx-bigquery\n\nJSON Expressions for BigQuery\n\n\n## Status\n\nMarch 2020 - Active but incomplete:  Can insert [tidy](https://en.wikipedia.org/wiki/Tidy_data) JSON documents into BigQuery while managing the schema.  Queries are not supported yet.\n\n## Overview\n\nThe library is intended to manage multiple BigQuery tables to give the illusion of one table with a dynamically managed schema. \n\n\n## Definitions\n\n* `partition` - Big data is split into separate containers based on age. This allows queries on recent data to use less resources, and allows old data to be dropped quickly\n* `cluster` - another name for the sorted order of the data in a partition. Sorting by the most common used lookup will make queries faster\n* `id` - The set of columns that identifies the document \n\n\n## Configuration\n\n* `table` - Any name you wish to give to this table series\n* `top_level_fields` - BigQuery demands that control columns are top-level.  Define them here.\n* `partition` - \n  * `field` - The dot-delimited field used to partition the tables (must be time datatype)\n  * `expire` - When BigQuery will automatically drop your data. \n* `id` - The identification of documents \n  * `field` - the set of columns to uniquely identify this document\n  * `version` - column used to determine age of a document; replacing newer with older\n* `cluster` - Columns used to sort the partitions \n* `schema` - {name: type} dictionary - needed when there is no data; BigQuery demands column definitions\n* `sharded` - *boolean* - set to `true` if you allow this library to track multiple tables. It allows for schema migration (expansion only), and for faster insert from a multitude of machines  \n* `account_info` - The information BigQuery provides to connect \n\n### Example\n\nThis is a complicated example. See [tests/config.json](https://github.com/klahnakoski/jx-bigquery/blob/dev/tests/config.json) for a minimal example.\n\n```json\n{\n    "table": "my_table_name",\n    "top_level_fields": {},\n    "partition": {\n        "field": "submit_time",\n        "expire": "2year"\n    },\n    "id": {\n        "field": "id",\n        "version": "last_modified"\n    },\n    "cluster": [\n        "id",\n        "last_modified"\n    ],\n    "schema": {\n        "id": "integer",\n        "submit_time": "time",\n        "last_modified": "time"\n    },\n    "sharded": true,\n    "account_info": {\n        "private_key_id": {\n            "$ref": "env://BIGQUERY_PRIVATE_KEY_ID"\n        },\n        "private_key": {\n            "$ref": "env://BIGQUERY_PRIVATE_KEY"\n        },\n        "type": "service_account",\n        "project_id": "my-project-id",\n        "client_email": "me@my_project.iam.gserviceaccount.com",\n        "client_id": "12345",\n        "auth_uri": "https://accounts.google.com/o/oauth2/auth",\n        "token_uri": "https://oauth2.googleapis.com/token",\n        "auth_provider_x509_cert_url": "https://www.googleapis.com/oauth2/v1/certs",\n        "client_x509_cert_url": "https://www.googleapis.com/robot/v1/metadata/x509/my-project.iam.gserviceaccount.com"\n    }\n}\n```\n\n## Usage\n\nSetup `Dataset` with an application name\n\n```python\n    dataset = bigquery.Dataset(\n        dataset=application_name, \n        kwargs=settings\n    )\n```\n\nCreate a table\n\n```python\n    destination = dataset.get_or_create_table(settings.destination)\n```\n\nInsert documents as you please\n\n\n```python\n    destination.extend(documents)\n```\n\nRequest a merge when done\n\n```python\n    destination.merge()\n```\n\n## Running tests\n\n\nFork and clone this repo. \n\n```\ngit clone https://github.com/klahnakoski/jx-bigquery.git \ncd jx-bigquery\npip install -r requirements.txt\n```\n\nYou will require a Google API key to run tests. The website will allow you to generate one and download a JSON file with the key.  Update the [tests/config.json](https://github.com/klahnakoski/jx-bigquery/blob/dev/tests/config.json) to point to that file: \n\n\n```\n# contents of tests/config.json\n{\n  "destination": {\n    "account_info": {\n      "$ref": "file:///e:/moz-fx-dev-ekyle-treeherder-a838a7718652.json"\n    }\n  },\n  "constants": {},\n  "debug": {\n    "trace": true\n  }\n}\n```\n\nThen you can run the tests\n\n```\npython -m unittest discover tests\n```\n\n> **NOTE** - the tests will create a `testing` dataset and generate/drop tables \n',
    long_description_content_type='text/markdown',
    name='jx-bigquery',
    packages=["jx_bigquery/expressions","jx_bigquery"],
    url='https://github.com/klahnakoski/jx-bigquery',
    version='3.61.20093'
)


# encoding: utf-8
#
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http:# mozilla.org/MPL/2.0/.
#
# Contact: Kyle Lahnakoski (kyle@lahnakoski.com)
#
from __future__ import absolute_import, division, unicode_literals

from jx_base.expressions import OrOp as OrOp_
from jx_bigquery.expressions import _utils
from jx_bigquery.expressions._utils import BQLang, check
from mo_dots import wrap
from mo_sql import SQL_OR, sql_iso, JoinSQL


class OrOp(OrOp_):
    @check
    def to_bq(self, schema, not_null=False, boolean=False):
        return wrap(
            [
                {
                    "name": ".",
                    "sql": {
                        "b": JoinSQL(
                            SQL_OR,
                            [
                                sql_iso(BQLang[t].to_bq(schema, boolean=True)[0].sql.b)
                                for t in self.terms
                            ],
                        )
                    },
                }
            ]
        )


_utils.OrOp = OrOp



# encoding: utf-8
#
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http:# mozilla.org/MPL/2.0/.
#
# Contact: Kyle Lahnakoski (kyle@lahnakoski.com)
#
from __future__ import absolute_import, division, unicode_literals

from jx_base.expressions import DivOp as DivOp_
from jx_bigquery.expressions._utils import BQLang, check
from mo_dots import Null, wrap
from mo_sql import sql_coalesce, sql_iso


class DivOp(DivOp_):
    @check
    def to_bq(self, schema, not_null=False, boolean=False):
        lhs = BQLang[self.lhs].to_bq(schema)[0].sql.n
        rhs = BQLang[self.rhs].to_bq(schema)[0].sql.n
        d = BQLang[self.default].to_bq(schema)[0].sql.n

        if lhs and rhs:
            if d == None:
                return wrap(
                    [{"name": ".", "sql": {"n": sql_iso(lhs) + " / " + sql_iso(rhs)}}]
                )
            else:
                return wrap(
                    [
                        {
                            "name": ".",
                            "sql": {
                                "n": sql_coalesce(
                                    [sql_iso(lhs) + " / " + sql_iso(rhs), d]
                                )
                            },
                        }
                    ]
                )
        else:
            return Null



# encoding: utf-8
#
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http:# mozilla.org/MPL/2.0/.
#
# Contact: Kyle Lahnakoski (kyle@lahnakoski.com)
#
from __future__ import absolute_import, division, unicode_literals

from jx_bigquery.sql import quote_column, GUID, ApiName

from jx_base.expressions import Variable as Variable_
from jx_base.queries import get_property_name
from jx_bigquery.expressions._utils import json_type_to_bq_type, check
from mo_dots import ROOT_PATH, relative_field, wrap
from mo_json import BOOLEAN, OBJECT
from mo_sql import SQL_IS_NOT_NULL, SQL_NULL, SQL_TRUE


class Variable(Variable_):
    @check
    def to_bq(self, schema, not_null=False, boolean=False, many=True):
        var_name = self.var
        if var_name == GUID:
            return wrap(
                [
                    {
                        "name": ".",
                        "sql": {"s": quote_column(ApiName(GUID))},
                        "nested_path": ROOT_PATH,
                    }
                ]
            )
        cols = schema.leaves(var_name)
        if not cols:
            # DOES NOT EXIST
            return wrap(
                [{"name": ".", "sql": {"0": SQL_NULL}, "nested_path": ROOT_PATH}]
            )
        acc = {}
        if boolean:
            for col in cols:
                cname = relative_field(col.name, var_name)
                nested_path = col.nested_path[0]
                if col.type == OBJECT:
                    value = SQL_TRUE
                elif col.type == BOOLEAN:
                    value = quote_column(col.es_column)
                else:
                    value = quote_column(col.es_column) + SQL_IS_NOT_NULL
                tempa = acc.setdefault(nested_path, {})
                tempb = tempa.setdefault(get_property_name(cname), {})
                tempb["b"] = value
        else:
            for col in cols:
                cname = relative_field(col.name, var_name)
                if col.jx_type == OBJECT:
                    prefix = self.var + "."
                    for cn, cs in schema.items():
                        if cn.startswith(prefix):
                            for child_col in cs:
                                tempa = acc.setdefault(child_col.nested_path[0], {})
                                tempb = tempa.setdefault(get_property_name(cname), {})
                                tempb[json_type_to_bq_type[col.type]] = quote_column(
                                    child_col.es_column
                                )
                else:
                    nested_path = col.nested_path[0]
                    tempa = acc.setdefault(nested_path, {})
                    tempb = tempa.setdefault(get_property_name(cname), {})
                    tempb[json_type_to_bq_type[col.jx_type]] = quote_column(
                        col.es_column
                    )

        return wrap(
            [
                {"name": cname, "sql": types, "nested_path": nested_path}
                for nested_path, pairs in acc.items()
                for cname, types in pairs.items()
            ]
        )



# encoding: utf-8
#
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http:# mozilla.org/MPL/2.0/.
#
# Contact: Kyle Lahnakoski (kyle@lahnakoski.com)
#
from __future__ import absolute_import, division, unicode_literals

from jx_base.expressions import FirstOp as FirstOp_
from jx_bigquery.expressions._utils import BQLang, check
from mo_logs import Log


class FirstOp(FirstOp_):
    @check
    def to_bq(self, schema, not_null=False, boolean=False):
        value = BQLang[self.term].to_bq(schema, not_null=True)
        for c in value:
            for t, v in c.sql.items():
                if t == "j":
                    Log.error("can not handle")
        return value



# encoding: utf-8
#
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http:# mozilla.org/MPL/2.0/.
#
# Contact: Kyle Lahnakoski (kyle@lahnakoski.com)
#
from __future__ import absolute_import, division, unicode_literals

from jx_base.expressions import BasicEqOp as BasicEqOp_
from jx_bigquery.expressions._utils import check
from mo_sql import SQL


class BasicEqOp(BasicEqOp_):
    @check
    def to_bq(self, schema, not_null=False, boolean=False, many=False):
        return (
            "("
            + SQL[self.rhs].to_bq(schema)
            + ") == ("
            + SQL[self.lhs].to_bq(schema)
            + ")"
        )
